/**
 * 
 */
package com.github.ansell.oas.data.impl;

import info.aduna.iteration.Iterations;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.atomic.AtomicInteger;

import org.openrdf.model.BNode;
import org.openrdf.model.Literal;
import org.openrdf.model.Model;
import org.openrdf.model.Namespace;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.ValueFactory;
import org.openrdf.model.impl.LinkedHashModel;
import org.openrdf.model.util.Namespaces;
import org.openrdf.model.vocabulary.OWL;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.RDFS;
import org.openrdf.model.vocabulary.XMLSchema;
import org.openrdf.query.BindingSet;
import org.openrdf.query.GraphQuery;
import org.openrdf.query.GraphQueryResult;
import org.openrdf.query.MalformedQueryException;
import org.openrdf.query.QueryEvaluationException;
import org.openrdf.query.QueryLanguage;
import org.openrdf.query.TupleQuery;
import org.openrdf.query.TupleQueryResult;
import org.openrdf.query.impl.DatasetImpl;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryConnection;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.RepositoryResult;
import org.openrdf.repository.sail.SailRepository;
import org.openrdf.rio.RDFFormat;
import org.openrdf.rio.RDFHandler;
import org.openrdf.rio.RDFHandlerException;
import org.openrdf.rio.RDFParseException;
import org.openrdf.rio.RDFWriter;
import org.openrdf.rio.Rio;
import org.openrdf.rio.UnsupportedRDFormatException;
import org.openrdf.rio.helpers.StatementCollector;
import org.openrdf.rio.ntriples.NTriplesUtil;
import org.openrdf.sail.Sail;
import org.openrdf.sail.memory.MemoryStore;
import org.restlet.data.Status;
import org.restlet.resource.ResourceException;
import org.restlet.security.User;
import org.semanticweb.owlapi.formats.OWLOntologyFormatFactory;
import org.semanticweb.owlapi.formats.OWLOntologyFormatFactoryRegistry;
import org.semanticweb.owlapi.formats.RDFXMLOntologyFormat;
import org.semanticweb.owlapi.formats.RioRDFOntologyFormat;
import org.semanticweb.owlapi.io.OWLOntologyDocumentSource;
import org.semanticweb.owlapi.io.OWLParser;
import org.semanticweb.owlapi.io.OWLParserException;
import org.semanticweb.owlapi.io.OWLParserFactory;
import org.semanticweb.owlapi.io.OWLParserFactoryRegistry;
import org.semanticweb.owlapi.io.RDFOntologyFormat;
import org.semanticweb.owlapi.io.ReaderDocumentSource;
import org.semanticweb.owlapi.io.StreamDocumentSource;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyChangeException;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyFormat;
import org.semanticweb.owlapi.model.OWLOntologyID;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.model.OWLOntologyManagerFactoryRegistry;
import org.semanticweb.owlapi.model.UnloadableImportException;
import org.semanticweb.owlapi.profiles.OWLProfile;
import org.semanticweb.owlapi.profiles.OWLProfileRegistry;
import org.semanticweb.owlapi.profiles.OWLProfileReport;
import org.semanticweb.owlapi.reasoner.InconsistentOntologyException;
import org.semanticweb.owlapi.reasoner.InferenceType;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.semanticweb.owlapi.reasoner.OWLReasonerFactory;
import org.semanticweb.owlapi.reasoner.OWLReasonerFactoryRegistry;
import org.semanticweb.owlapi.reasoner.ReasonerInterruptedException;
import org.semanticweb.owlapi.reasoner.TimeOutException;
import org.semanticweb.owlapi.rio.RioRenderer;
import org.semanticweb.owlapi.util.InferredOntologyGenerator;
import org.semanticweb.owlapi.vocab.PrefixOWLOntologyFormat;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.data.DataHandler;
import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.objects.ResourceCount;
import com.github.ansell.oas.objects.SearchResult;
import com.github.ansell.oas.objects.impl.AnnotationImpl;
import com.github.ansell.oas.objects.impl.SearchResultImpl;
import com.github.ansell.oas.objects.impl.UriCountImpl;
import com.github.ansell.oas.utils.OasConstants;
import com.github.ansell.oas.utils.SesameUtils;
import com.github.ansell.oas.utils.ValueComparator;
import com.github.ansell.oas.webservice.OasAction;

/**
 * A data handler implementation that creates a temporary in memory RDF database for testing
 * purposes unless it is given one when it is created.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class SesameDataHandlerImpl implements DataHandler
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private Repository repository;
    
    private final OWLOntologyManager manager;
    
    /**
     * Default constructor, creates a new in memory store.
     */
    public SesameDataHandlerImpl()
    {
        try
        {
            this.repository = new SailRepository(new MemoryStore());
            
            this.repository.initialize();
        }
        catch(final RepositoryException ex)
        {
            this.log.error("Exception while creating in memory repository.");
            throw new RuntimeException(ex);
        }
        
        this.manager = OWLOntologyManagerFactoryRegistry.createOWLOntologyManager();
    }
    
    /**
     * Constructor to define the pre-existing repository to store and retrieve information from.
     * 
     * NOTE: the repository must be initialized prior to use!
     * 
     * @param nextRepository
     *            The repository to store and retrieve data from.
     */
    public SesameDataHandlerImpl(final Repository nextRepository)
    {
        this.repository = nextRepository;
        
        this.manager = OWLOntologyManagerFactoryRegistry.createOWLOntologyManager();
    }
    
    /**
     * Constructor to define the pre-existing repository to store and retrieve information from.
     * 
     * NOTE: the repository must be initialized prior to use!
     * 
     * @param nextRepository
     *            The repository to store and retrieve data from.
     */
    public SesameDataHandlerImpl(final Sail nextRepository)
    {
        this(new SailRepository(nextRepository));
    }
    
    @Override
    public void addLogDetailsForRequest(final String message, final String referenceUri, final OasAction action,
            final User currentUser, final String ipAddress)
    {
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            // DebugUtils.printContexts(conn);
            // DebugUtils.printContents(conn, SesameUtils.AUDIT_LOG_GRAPH);
            
            final ValueFactory vf = conn.getValueFactory();
            
            conn.begin();
            final URI recordId =
                    vf.createURI(OasConstants.TAGINSTANCE_NS + ":auditlog:" + UUID.randomUUID().toString());
            
            conn.add(recordId, RDF.TYPE, OasConstants.URN_TAG_AUDITLOG_RECORD, SesameUtils.AUDIT_LOG_GRAPH);
            
            conn.add(recordId, RDFS.COMMENT, vf.createLiteral(message, XMLSchema.STRING), SesameUtils.AUDIT_LOG_GRAPH);
            conn.add(recordId, OasConstants.DC_TERMS_MODIFIED, vf.createLiteral(new Date()),
                    SesameUtils.AUDIT_LOG_GRAPH);
            
            if(action != null)
            {
                conn.add(recordId, OasConstants.URN_TAG_LOGACTION,
                        vf.createLiteral(action.toString(), XMLSchema.STRING), SesameUtils.AUDIT_LOG_GRAPH);
                conn.add(recordId, OasConstants.URN_TAG_LOGOPERATION_VERB, action.getMethod().getUri(),
                        SesameUtils.AUDIT_LOG_GRAPH);
            }
            
            if(currentUser != null)
            {
                conn.add(recordId, OasConstants.FOAF_MAKER,
                        vf.createLiteral(currentUser.getIdentifier(), XMLSchema.STRING), SesameUtils.AUDIT_LOG_GRAPH);
            }
            
            if(referenceUri != null && !referenceUri.trim().isEmpty())
            {
                conn.add(recordId, OasConstants.URN_TAG_AUDITLOG_REF, vf.createLiteral(referenceUri, XMLSchema.STRING),
                        SesameUtils.AUDIT_LOG_GRAPH);
            }
            
            if(ipAddress != null && !ipAddress.trim().isEmpty())
            {
                conn.add(recordId, OasConstants.URN_TAG_USERSIPADDRESS, vf.createLiteral(ipAddress, XMLSchema.STRING),
                        SesameUtils.AUDIT_LOG_GRAPH);
            }
            
            conn.commit();
        }
        catch(final RepositoryException re)
        {
            // Something went wrong during the transaction, so we roll it back
            try
            {
                if(conn != null && conn.isActive())
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to rollback the transaction", ex);
            }
            
            // Do not throw an exception when logging fails, rely on the log for that
            // throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
            // "Failed to complete audit log storage, as a repository exception occurred.", re);
            this.log.error("ERROR: Failed to complete audit log storage, as a repository exception occurred.", re);
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to rollback the transaction", ex);
            }
        }
        
    }
    
    @Override
    public BNode createBlankNode()
    {
        return this.repository.getValueFactory().createBNode();
    }
    
    @Override
    public Literal createLiteral(final String plainLiteralString)
    {
        return this.repository.getValueFactory().createLiteral(plainLiteralString);
    }
    
    @Override
    public Literal createLiteral(final String languageLiteralString, final String language)
    {
        return this.repository.getValueFactory().createLiteral(languageLiteralString, language);
    }
    
    @Override
    public Literal createLiteral(final String typedLiteralString, final URI datatypeUri)
    {
        return this.repository.getValueFactory().createLiteral(typedLiteralString, datatypeUri);
    }
    
    @Override
    public URI createURI(final String uri)
    {
        return this.repository.getValueFactory().createURI(uri);
    }
    
    @Override
    public boolean deleteAnnotation(final URI annotationIdUri)
    {
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            // do a quick check to see if there were any statements in a graph with that identifier
            if(conn.hasStatement((URI)null, (URI)null, (Value)null, false, annotationIdUri))
            {
                conn.begin();
                
                // Annotations are stored in a graph based on their identifying URI
                conn.remove((URI)null, null, null, annotationIdUri);
                
                // If everything went as planned, we can commit the result
                conn.commit();
                
                return true;
            }
            else
            {
                this.log.debug("Could not delete annotation: {}", annotationIdUri.stringValue());
                // if there were no statements in a graph with that identifier, return false
                return false;
            }
        }
        catch(final RepositoryException re)
        {
            // Something went wrong during the transaction, so we roll it back
            try
            {
                if(conn != null && conn.isActive())
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to rollback the transaction", ex);
            }
            
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete storage, as a repository exception occurred.", re);
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to rollback the transaction", ex);
            }
        }
    }
    
    @Override
    public boolean deleteOntologyAllVersions(final URI nextOntologyUri)
    {
        if(nextOntologyUri == null)
        {
            throw new IllegalArgumentException("Cannot delete a null ontology");
        }
        
        RepositoryConnection conn = null;
        
        // this.addLogDetailsForRequest(annotationIdUri.stringValue(), request,
        // DataHandlerConstants.DELETE);
        try
        {
            conn = this.repository.getConnection();
            
            conn.begin();
            
            // TODO FIXME: Do an authentication check to see whether the logged in user can delete
            // this ontology
            
            if(conn.hasStatement(nextOntologyUri, RDF.TYPE, OWL.ONTOLOGY, false, SesameUtils.ONTOLOGY_MANAGER_GRAPH))
            {
                // remove the current version annotation
                conn.remove(nextOntologyUri, OasConstants.OMV_CURRENTVERSION, null, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                
                RepositoryResult<Statement> asList2 = null;
                
                try
                {
                    asList2 =
                            conn.getStatements(nextOntologyUri, OasConstants.OWL_VERSIONIRI, null, false,
                                    SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    
                    synchronized(this.manager)
                    {
                        // clear the graphs/contexts for each of the versions
                        while(asList2.hasNext())
                        {
                            final Statement nextStatement = asList2.next();
                            conn.remove((URI)null, null, null, (URI)nextStatement.getObject());
                            this.manager.removeOntology(new OWLOntologyID(nextOntologyUri, (URI)nextStatement
                                    .getObject()));
                        }
                        
                        this.manager.removeOntology(new OWLOntologyID(IRI.create(nextOntologyUri)));
                    }
                    
                    conn.remove(nextOntologyUri, OasConstants.OWL_VERSIONIRI, null, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    
                    conn.remove(nextOntologyUri, OWL.PRIORVERSION, null, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    conn.remove(nextOntologyUri, RDF.TYPE, OWL.ONTOLOGY, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                }
                finally
                {
                    if(asList2 != null)
                    {
                        asList2.close();
                    }
                }
            }
            else
            {
                conn.rollback();
                
                // we are not managing any ontologies identified by this URI
                return false;
            }
            
            // If everything went as planned, we can commit the result
            conn.commit();
            
            return true;
        }
        catch(final RepositoryException re)
        {
            // Something went wrong during the transaction, so we roll it back
            try
            {
                if(conn != null && conn.isActive())
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to rollback the transaction", ex);
            }
            
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete storage, as a repository exception occurred.", re);
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to close the connection", ex);
            }
        }
    }
    
    @Override
    public boolean deleteOntologyVersion(final Ontology nextOntology)
    {
        if(nextOntology == null)
        {
            throw new IllegalArgumentException("Cannot delete a null ontology");
        }
        
        if(this.log.isDebugEnabled())
        {
            this.log.info("before deleteOntologyVersion nextOntology.getOntologyUri()={}",
                    nextOntology.getOntologyUri());
            this.log.info("before deleteOntologyVersion nextOntology.getOntologyVersionUri()={}",
                    nextOntology.getOntologyVersionUri());
        }
        
        // Attach a version IRI to the given ontology if it was not given originally and one
        // was discovered
        final boolean checkVersionUriUniqueness =
                nextOntology.ensureVersionUriUniqueness(this.repository, false, true,
                        SesameUtils.ONTOLOGY_MANAGER_GRAPH);
        
        if(this.log.isDebugEnabled())
        {
            this.log.debug("before deleteOntologyVersion managedOntologyVersionUri={}",
                    nextOntology.getOntologyVersionUri());
        }
        
        // FIXME: The following may be incorrect without managedOntologyVersionUri
        if(nextOntology.getOntologyVersionUri() == null)
        {
            throw new IllegalArgumentException("Cannot delete this version of the ontology as it was not managed");
        }
        
        synchronized(this.manager)
        {
            this.manager.removeOntology(nextOntology.toOWLOntologyID());
        }
        
        RepositoryConnection conn = null;
        
        // this.addLogDetailsForRequest(annotationIdUri.stringValue(), request,
        // DataHandlerConstants.DELETE);
        
        final Map<URI, Collection<URI>> children =
                this.getChildren(nextOntology, OasConstants.OWL_THING, true, 100000, 0);
        
        // We must not allow deletion of any ontologies that have annotations referencing terms
        // TODO: Check with other versions of this ontology to verify whether there are other
        // ontologies which may also provide the term
        for(final URI nextChildUri : children.keySet())
        {
            // check the child URI, as long as it is not OWL_THING
            if(!nextChildUri.equals(OasConstants.OWL_THING))
            {
                final ResourceCount annotationCountsByOntologyTermUri =
                        this.getAnnotationCountsByOntologyTermUri(nextChildUri, null, false);
                
                if(annotationCountsByOntologyTermUri.getCount() > 0)
                {
                    throw new ResourceException(
                            Status.CLIENT_ERROR_CONFLICT,
                            "Cannot delete the ontology as there are annotations which depend on a term in the ontology: Example term URI that has annotations is: term="
                                    + nextChildUri);
                }
            }
            
            for(final URI nextDeeperChildUri : children.get(nextChildUri))
            {
                if(!nextDeeperChildUri.equals(OasConstants.OWL_THING))
                {
                    final ResourceCount annotationCountsByOntologyTermUri =
                            this.getAnnotationCountsByOntologyTermUri(nextDeeperChildUri, null, false);
                    
                    if(annotationCountsByOntologyTermUri.getCount() > 0)
                    {
                        throw new ResourceException(
                                Status.CLIENT_ERROR_CONFLICT,
                                "Cannot delete the ontology as there are annotations which depend on a term in the ontology: Example term URI that has annotations is: term="
                                        + nextDeeperChildUri);
                    }
                }
            }
        }
        
        try
        {
            conn = this.repository.getConnection();
            
            // NOTE: Do NOT enable this in production, only enable for debugging
            if(this.log.isTraceEnabled())
            {
                this.log.trace("before deleteOntologyVersion conn.size()=" + conn.size());
                this.log.trace("before deleteOntologyVersion conn.size(ontologyVersionUri)="
                        + conn.size(nextOntology.getOntologyVersionUri()));
                this.log.trace("before deleteOntologyVersion conn.size(UriConstants.ONTOLOGY_MANAGER_GRAPH)="
                        + conn.size(SesameUtils.ONTOLOGY_MANAGER_GRAPH));
                
                for(final Statement nextStatement : Iterations.asList(conn.getStatements(null, null, null, false)))
                {
                    this.log.trace("allContexts nextStatement=" + nextStatement);
                }
                
                this.log.trace("");
                
                for(final Statement nextStatement : Iterations.asList(conn.getStatements(null, null, null, false,
                        SesameUtils.ONTOLOGY_MANAGER_GRAPH)))
                {
                    this.log.trace("ontologyManagerGraph nextStatement=" + nextStatement);
                }
            }
            
            conn.begin();
            
            // TODO: Do an authentication check to see whether the logged in user can delete this
            // particular ontology if ontologies have registered maintainers in the future
            
            // check if we know about the existence of the ontology
            if(conn.hasStatement(nextOntology.getOntologyUri(), RDF.TYPE, OWL.ONTOLOGY, false,
                    SesameUtils.ONTOLOGY_MANAGER_GRAPH))
            {
                // check if we know about the existence of the given version of the ontology
                // NOTE: Once we have a managed ontology version URI it is unlikely that this step
                // should fail.
                if(conn.hasStatement(nextOntology.getOntologyUri(), OasConstants.OWL_VERSIONIRI,
                        nextOntology.getOntologyVersionUri(), false, SesameUtils.ONTOLOGY_MANAGER_GRAPH))
                {
                    // remove the ontology version from the management graph
                    conn.remove(nextOntology.getOntologyUri(), OasConstants.OWL_VERSIONIRI,
                            nextOntology.getOntologyVersionUri(), SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    
                    conn.remove(nextOntology.getOntologyUri(), OWL.PRIORVERSION, nextOntology.getOntologyVersionUri(),
                            SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    
                    // TODO: If we store information about the inferred ontology, need to remove it
                    // before doing this remove
                    conn.remove(nextOntology.getOntologyVersionUri(), null, null, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    
                    if(conn.hasStatement(nextOntology.getOntologyUri(), OasConstants.OAS_CURRENTINFERREDAXIOMSONTOLOGY,
                            nextOntology.getOntologyVersionUri(), false, SesameUtils.ONTOLOGY_MANAGER_GRAPH))
                    {
                        conn.remove(nextOntology.getOntologyUri(), OasConstants.OAS_CURRENTINFERREDAXIOMSONTOLOGY,
                                nextOntology.getOntologyVersionUri(), SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    }
                    
                    // if the version we are deleting was the current version, we need to try to
                    // reset the current version to the next most recent, in our wisdom, ontology
                    // version
                    if(conn.hasStatement(nextOntology.getOntologyUri(), OasConstants.OMV_CURRENTVERSION,
                            nextOntology.getOntologyVersionUri(), false, SesameUtils.ONTOLOGY_MANAGER_GRAPH))
                    {
                        conn.remove(nextOntology.getOntologyUri(), OasConstants.OMV_CURRENTVERSION,
                                nextOntology.getOntologyVersionUri(), SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                        
                        // Try to negotiate another current version using best effort
                        RepositoryResult<Statement> asList = null;
                        try
                        {
                            asList =
                                    conn.getStatements(nextOntology.getOntologyUri(), OasConstants.OWL_VERSIONIRI,
                                            null, false, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                            
                            // if there are not any versions, then delete the record of the ontology
                            if(!asList.hasNext())
                            {
                                // there were no other versions known for this ontology, so remove
                                // all recognition of the ontology from the ontology manager graph
                                conn.remove(nextOntology.getOntologyUri(), null, null,
                                        SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                            }
                            else
                            {
                                // search through the remaining versions and try to pick the best
                                // one based on prior version links
                                int maxCount = -1;
                                URI maxContext = null;
                                
                                while(asList.hasNext())
                                {
                                    final Statement nextStatement = asList.next();
                                    
                                    if(nextStatement.getObject() instanceof URI)
                                    {
                                        final URI nextContext = (URI)nextStatement.getObject();
                                        
                                        // find out how many prior versions were known when the next
                                        // version was uploaded, to determine the order
                                        // NOTE: This assumes that all existing priorVersion
                                        // statements have been created using this algorithm
                                        // In OAS, this assumes that any preexisting priorVersion
                                        // statements were wiped after uploading and replaced using
                                        // the current knowledge
                                        final List<Statement> asList2 =
                                                Iterations.asList(conn.getStatements(nextOntology.getOntologyUri(),
                                                        OWL.PRIORVERSION, null, false,
                                                        nextOntology.getOntologyVersionUri()));
                                        
                                        if(maxContext == null || asList2.size() > maxCount)
                                        {
                                            maxContext = nextContext;
                                            maxCount = asList2.size();
                                        }
                                    }
                                    else
                                    {
                                        this.log.error("Found an ontology version that was not a URI nextStatement.getObject()="
                                                + nextStatement.getObject().stringValue());
                                    }
                                }
                                
                                this.log.debug("Max count was maxCount={} maxContext={}", maxCount, maxContext);
                                
                                // setup the new current version if we discovered one
                                if(maxContext != null)
                                {
                                    conn.add(nextOntology.getOntologyUri(), OasConstants.OMV_CURRENTVERSION,
                                            maxContext, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                                    
                                    // If we have stored an inferred axioms ontology for the version
                                    // we are reverting to then replace the current reference with
                                    // it
                                    final RepositoryResult<Statement> inferredAxiomsOntology =
                                            conn.getStatements(maxContext, OasConstants.OAS_INFERREDAXIOMSONTOLOGY,
                                                    null, false, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                                    
                                    try
                                    {
                                        if(inferredAxiomsOntology.hasNext())
                                        {
                                            conn.add(nextOntology.getOntologyUri(),
                                                    OasConstants.OAS_CURRENTINFERREDAXIOMSONTOLOGY,
                                                    inferredAxiomsOntology.next().getObject(),
                                                    SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                                        }
                                    }
                                    finally
                                    {
                                        inferredAxiomsOntology.close();
                                    }
                                }
                            }
                        }
                        finally
                        {
                            if(asList != null)
                            {
                                asList.close();
                            }
                        }
                    }
                    
                    // remove the context for the managed ontology version URI that is being deleted
                    conn.remove((URI)null, null, null, nextOntology.getOntologyVersionUri());
                }
                else
                {
                    conn.rollback();
                    
                    this.log.debug("Returning false as we didn't know about this version of the ontology");
                    // we don't know about the given version of the ontology so we return false to
                    // indicate that we didn't make any changes
                    return false;
                }
            }
            else
            {
                conn.rollback();
                
                this.log.debug("Returning false as we didn't know about the existence of the ontology");
                // we are not managing any ontologies identified by this URI
                return false;
            }
            
            // If everything went as planned, we can commit the result
            conn.commit();
            
            return true;
        }
        catch(final RepositoryException re)
        {
            // Something went wrong during the transaction, so we roll it back
            try
            {
                if(conn != null)
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to rollback the transaction", ex);
            }
            
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete storage, as a repository exception occurred.", re);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
    }
    
    @Override
    public void downloadOntology(final Ontology nextOntology, final OutputStream outputStream, final String mimeType)
    {
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            // the RDFFormat to export to, automatically using RDF/XML as fallback if we didn't
            // recognise what they sent
            final RDFFormat rdfFormat = RDFFormat.forMIMEType(mimeType, RDFFormat.RDFXML);
            
            // the writer to export to
            final RDFWriter createWriter = Rio.createWriter(rdfFormat, outputStream);
            
            // the context to export
            // final URI managedOntologyVersionUri =
            // nextOntology.getManagedOntologyVersionUri(this.repository,
            // UriConstants.ONTOLOGY_MANAGER_GRAPH);
            
            // export all statements relating to this version of the Ontology
            // conn.exportStatements(managedOntologyVersionUri, null, null, true, createWriter,
            // UriConstants.ONTOLOGY_MANAGER_GRAPH);
            
            // export all statements relating to the overall ontologyUri
            // conn.exportStatements(nextOntology.getOntologyUri(), null, null, true, createWriter,
            // UriConstants.ONTOLOGY_MANAGER_GRAPH);
            
            // then export the actual contents of the ontology
            createWriter.startRDF();
            
            final RepositoryResult<Namespace> namespaces = conn.getNamespaces();
            
            try
            {
                while(namespaces.hasNext())
                {
                    final Namespace nextNamespace = namespaces.next();
                    createWriter.handleNamespace(nextNamespace.getPrefix(), nextNamespace.getName());
                }
            }
            finally
            {
                namespaces.close();
            }
            
            // then export the actual contents of the ontology
            final RepositoryResult<Statement> concreteStatements =
                    conn.getStatements(null, null, null, true, nextOntology.getOntologyVersionUri());
            
            try
            {
                while(concreteStatements.hasNext())
                {
                    final Statement nextConcreteStatement = concreteStatements.next();
                    createWriter.handleStatement(nextConcreteStatement);
                }
            }
            finally
            {
                concreteStatements.close();
            }
            
            createWriter.endRDF();
        }
        catch(final RepositoryException re)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download, as a repository exception occurred.", re);
        }
        catch(final RDFHandlerException rdfhe)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download as an rdf handler exception occurred.", rdfhe);
        }
        catch(final UnsupportedRDFormatException urdffe)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download as an unsupported RDF format exception occurred.", urdffe);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
    }
    
    @Override
    public void downloadOntologySubset(final Ontology nextOntology, final ByteArrayOutputStream outputStream,
            final String mimeType, final URI parentUri, final boolean recursive, final boolean addCounts)
    {
        URI actualParent = parentUri;
        
        if(actualParent == null)
        {
            actualParent = OWL.THING;
        }
        
        final Ontology versionedOntology =
                this.getVersionedOntology(nextOntology.getOntologyUri(), nextOntology.getOntologyVersionUri());
        
        if(versionedOntology == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find requested ontology version");
        }
        
        final DatasetImpl queryDataset = new DatasetImpl();
        queryDataset.addDefaultGraph(versionedOntology.getOntologyVersionUri());
        queryDataset.addDefaultGraph(versionedOntology.getInferredOntologyUri());
        
        final StringBuilder queryString = new StringBuilder(256);
        
        // select all distinct parent/child combinations using the query that fits this ontologies
        // structure
        queryString.append("SELECT DISTINCT ?parent ");
        queryString.append(" ?child ");
        queryString.append(" WHERE { ");
        
        if(recursive)
        {
            // this.log.info("performing recursive query for ontology subset");
            queryString.append(nextOntology.getRecursiveChildQuery());
        }
        else
        {
            // this.log.info("performing direct query for ontology subset");
            queryString.append(nextOntology.getDirectChildQuery());
        }
        
        queryString.append(" } ");
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            final ValueFactory vf = conn.getValueFactory();
            // this.log.info("query: {}", queryString);
            
            final TupleQuery tupleQuery = conn.prepareTupleQuery(QueryLanguage.SPARQL, queryString.toString());
            
            tupleQuery.setDataset(queryDataset);
            
            tupleQuery.clearBindings();
            
            tupleQuery.setBinding("parent", actualParent);
            
            // the RDFFormat to export to, automatically using RDF/XML as fallback if we didn't
            // recognise what they sent
            final RDFFormat rdfFormat = RDFFormat.forMIMEType(mimeType, RDFFormat.RDFXML);
            
            // the writer to export to
            final RDFWriter writer = Rio.createWriter(rdfFormat, outputStream);
            
            final TupleQueryResult evaluate = tupleQuery.evaluate();
            
            writer.startRDF();
            
            List<URI> children = new ArrayList<URI>();
            
            while(evaluate.hasNext())
            {
                final BindingSet next = evaluate.next();
                
                URI parent = null;
                URI child = null;
                
                if(!next.hasBinding("parent"))
                {
                    throw new RuntimeException("Parent was not bound");
                }
                
                if(!next.hasBinding("child"))
                {
                    throw new RuntimeException("Child was not bound, parent: "
                            + next.getBinding("parent").getValue().stringValue());
                }
                
                if(next.getBinding("parent").getValue() instanceof URI)
                {
                    parent = (URI)next.getBinding("parent").getValue();
                }
                else
                {
                    throw new RuntimeException("Parent was not a URI: " + next.getBinding("parent"));
                }
                
                if(next.getBinding("child").getValue() instanceof URI)
                {
                    child = (URI)next.getBinding("child").getValue();
                    children.add(child);
                }
                else
                {
                    throw new RuntimeException("Child was not a URI: " + next.getBinding("child"));
                }
                
                final Statement statement = vf.createStatement(parent, RDFS.SEEALSO, child);
                
                // this.log.trace("statement: {}", statement);
                
                writer.handleStatement(statement);
                
                // get labels for the child
                final List<Literal> labelsForUri = this.getLabelsForUri(versionedOntology, child);
                
                for(final Literal nextLabel : labelsForUri)
                {
                    writer.handleStatement(vf.createStatement(child, RDFS.LABEL, nextLabel));
                }
            }
            
            if(addCounts)
            {
                final ConcurrentMap<URI, ResourceCount> directParentCount =
                        this.getChildrenCount(versionedOntology, Arrays.asList(actualParent), false);
                
                mapCounts(writer, directParentCount, OasConstants.OAS_URI_DIRECT_COUNT, vf);
                
                final ConcurrentMap<URI, ResourceCount> recursiveParentCount =
                        this.getChildrenCount(versionedOntology, Arrays.asList(actualParent), true);
                
                mapCounts(writer, recursiveParentCount, OasConstants.OAS_URI_RECURSIVE_COUNT, vf);
                
                // TODO: Reenable this check
                // if(recursiveParentCount < directParentCount)
                // {
                // this.log.error("Found recursive count that was less than the direct count {}",
                // actualParent.stringValue());
                // }
                
                final ConcurrentMap<URI, ResourceCount> directChildrenCounts =
                        this.getChildrenCount(versionedOntology, children, false);
                
                mapCounts(writer, directChildrenCounts, OasConstants.OAS_URI_DIRECT_COUNT, vf);
                
                final ConcurrentMap<URI, ResourceCount> recursiveChildrenCounts =
                        this.getChildrenCount(versionedOntology, children, true);
                
                mapCounts(writer, recursiveChildrenCounts, OasConstants.OAS_URI_RECURSIVE_COUNT, vf);
                
                // TODO: Reenable this check
                // if(recursiveChildCount < directChildCount)
                // {
                // this.log.error("Found recursive count that was less than the direct count {}",
                // child.stringValue());
                // }
            }
            writer.endRDF();
        }
        catch(final RepositoryException re)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download, as a repository exception occurred.", re);
        }
        catch(final RDFHandlerException rdfhe)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download as an rdf handler exception occurred.", rdfhe);
        }
        catch(final UnsupportedRDFormatException urdffe)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download as an unsupported RDF format exception occurred.", urdffe);
        }
        catch(final MalformedQueryException mqe)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download as a malformed query exception occurred.", mqe);
        }
        catch(final QueryEvaluationException qee)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download as a query evaluation exception occurred.", qee);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
    }
    
    private void mapCounts(RDFHandler handler, ConcurrentMap<URI, ResourceCount> counts, URI predicateUri,
            ValueFactory vf) throws RDFHandlerException
    {
        for(URI nextURI : counts.keySet())
        {
            handler.handleStatement(vf.createStatement(nextURI, predicateUri,
                    vf.createLiteral(counts.get(nextURI).getCount())));
        }
    }
    
    @Override
    public void downloadOntologyWithInferences(final Ontology nextOntology, final OutputStream outputStream,
            final String mimeType)
    {
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            // the RDFFormat to export to, automatically using RDF/XML as fallback if we didn't
            // recognise what they sent
            final RDFFormat rdfFormat = RDFFormat.forMIMEType(mimeType, RDFFormat.RDFXML);
            
            // the writer to export to
            final RDFWriter handler = Rio.createWriter(rdfFormat, outputStream);
            
            // the context to export
            // final URI managedOntologyVersionUri =
            // nextOntology.getManagedOntologyVersionUri(this.repository,
            // UriConstants.ONTOLOGY_MANAGER_GRAPH);
            
            // export all statements relating to this version of the Ontology
            // conn.exportStatements(managedOntologyVersionUri, null, null, true, createWriter,
            // UriConstants.ONTOLOGY_MANAGER_GRAPH);
            
            // export all statements relating to the overall ontologyUri
            // conn.exportStatements(nextOntology.getOntologyUri(), null, null, true, createWriter,
            // UriConstants.ONTOLOGY_MANAGER_GRAPH);
            
            handler.startRDF();
            
            final RepositoryResult<Namespace> namespaces = conn.getNamespaces();
            
            try
            {
                while(namespaces.hasNext())
                {
                    final Namespace nextNamespace = namespaces.next();
                    handler.handleNamespace(nextNamespace.getPrefix(), nextNamespace.getName());
                }
            }
            finally
            {
                namespaces.close();
            }
            
            final RepositoryResult<Statement> managedOntologyManagerStatements =
                    conn.getStatements(null, null, nextOntology.getOntologyVersionUri(), true,
                            SesameUtils.ONTOLOGY_MANAGER_GRAPH);
            
            try
            {
                while(managedOntologyManagerStatements.hasNext())
                {
                    final Statement nextOntologyManagerStatement = managedOntologyManagerStatements.next();
                    handler.handleStatement(nextOntologyManagerStatement);
                }
            }
            finally
            {
                managedOntologyManagerStatements.close();
            }
            
            final RepositoryResult<Statement> ontologyManagerStatements =
                    conn.getStatements(nextOntology.getOntologyVersionUri(), null, null, true,
                            SesameUtils.ONTOLOGY_MANAGER_GRAPH);
            
            try
            {
                while(ontologyManagerStatements.hasNext())
                {
                    final Statement nextOntologyManagerStatement = ontologyManagerStatements.next();
                    handler.handleStatement(nextOntologyManagerStatement);
                }
            }
            finally
            {
                ontologyManagerStatements.close();
            }
            
            final RepositoryResult<Statement> inferredManagerStatements =
                    conn.getStatements(nextOntology.getInferredOntologyUri(), null, null, true,
                            SesameUtils.ONTOLOGY_MANAGER_GRAPH);
            
            try
            {
                while(inferredManagerStatements.hasNext())
                {
                    final Statement nextInferredManagerStatement = inferredManagerStatements.next();
                    handler.handleStatement(nextInferredManagerStatement);
                }
            }
            finally
            {
                inferredManagerStatements.close();
            }
            
            // then export the actual contents of the ontology
            final RepositoryResult<Statement> concreteStatements =
                    conn.getStatements(null, null, null, true, nextOntology.getOntologyVersionUri());
            
            try
            {
                while(concreteStatements.hasNext())
                {
                    final Statement nextConcreteStatement = concreteStatements.next();
                    handler.handleStatement(nextConcreteStatement);
                }
            }
            finally
            {
                concreteStatements.close();
            }
            
            final RepositoryResult<Statement> inferredStatements =
                    conn.getStatements(null, null, null, true, nextOntology.getInferredOntologyUri());
            
            try
            {
                while(inferredStatements.hasNext())
                {
                    final Statement nextInferredStatement = inferredStatements.next();
                    handler.handleStatement(nextInferredStatement);
                }
            }
            finally
            {
                inferredStatements.close();
            }
            
            handler.endRDF();
        }
        catch(final RepositoryException re)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download, as a repository exception occurred.", re);
        }
        catch(final RDFHandlerException rdfhe)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download as an rdf handler exception occurred.", rdfhe);
        }
        catch(final UnsupportedRDFormatException urdffe)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete download as an unsupported RDF format exception occurred.", urdffe);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
    }
    
    private Collection<Statement> executeSparqlConstructQuery(final String sparqlQuery) throws MalformedQueryException,
        QueryEvaluationException, RepositoryException
    {
        final Collection<Statement> allResults = new ArrayList<Statement>();
        
        RepositoryConnection conn = null;
        try
        {
            conn = this.repository.getConnection();
            
            final GraphQuery query = conn.prepareGraphQuery(QueryLanguage.SPARQL, sparqlQuery);
            final GraphQueryResult result = query.evaluate();
            
            while(result.hasNext())
            {
                allResults.add(result.next());
            }
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
        
        return allResults;
        
    }
    
    private List<Literal> executeSparqlSelectQueryOnSingleLiteralVariable(final String sparqlQuery,
            final String variableName) throws QueryEvaluationException, RepositoryException, MalformedQueryException
    {
        final List<Literal> allResults = new ArrayList<Literal>();
        
        RepositoryConnection conn = null;
        try
        {
            conn = this.repository.getConnection();
            
            final TupleQuery query = conn.prepareTupleQuery(QueryLanguage.SPARQL, sparqlQuery);
            final TupleQueryResult result = query.evaluate();
            
            while(result.hasNext())
            {
                final BindingSet nextResult = result.next();
                
                allResults.add((Literal)nextResult.getBinding(variableName).getValue());
            }
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
        return allResults;
    }
    
    private List<URI> executeSparqlSelectQueryOnSingleUriVariable(final String sparqlQuery, final String variableName)
        throws RepositoryException, MalformedQueryException, QueryEvaluationException
    {
        final List<URI> allResults = new ArrayList<URI>();
        
        RepositoryConnection conn = null;
        try
        {
            conn = this.repository.getConnection();
            
            final TupleQuery query = conn.prepareTupleQuery(QueryLanguage.SPARQL, sparqlQuery);
            final TupleQueryResult result = query.evaluate();
            
            while(result.hasNext())
            {
                final BindingSet nextResult = result.next();
                
                allResults.add((URI)nextResult.getBinding(variableName).getValue());
            }
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
        return allResults;
    }
    
    @Override
    public Annotation getAnnotation(final URI annotationIdUri)
    {
        try
        {
            return SesameUtils.createAnnotationByUriFromRepository(annotationIdUri, this.repository,
                    SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to get requested annotation from repository because of a repository exception", ex);
        }
    }
    
    @Override
    public ResourceCount getAnnotationCountsByAnnotatedObjectUri(final URI objectUriResource)
    {
        final ResourceCount result = new UriCountImpl(0);
        final StringBuilder selectQuery = new StringBuilder(200);
        
        // Count all of the annotations directly referencing the given ontology term uri
        selectQuery.append(" SELECT (COUNT(DISTINCT ?annotationUri) AS ?countAnnotations) ");
        selectQuery.append(" WHERE { ");
        selectQuery.append(" GRAPH <");
        selectQuery.append(SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
        selectQuery.append("> { ");
        selectQuery.append(" ?annotationUri <");
        selectQuery.append(RDF.TYPE.stringValue());
        selectQuery.append("> <");
        selectQuery.append(OasConstants.TAGGING_TAGGING);
        selectQuery.append("> . ");
        selectQuery.append(" }");
        selectQuery.append(" GRAPH ?annotationUri { ");
        selectQuery.append(" ?annotationUri <");
        selectQuery.append(OasConstants.MOAT_TAG_MEANING.stringValue());
        selectQuery.append("> ?ontologyTermUri . ");
        selectQuery.append(" ?annotationUri <");
        selectQuery.append(OasConstants.TAGGING_TAGGED_RESOURCE.stringValue());
        selectQuery.append("> <");
        selectQuery.append(objectUriResource.stringValue());
        selectQuery.append("> . ");
        selectQuery.append(" } ");
        selectQuery.append(" }");
        
        try
        {
            final List<Literal> list =
                    this.executeSparqlSelectQueryOnSingleLiteralVariable(selectQuery.toString(), "countAnnotations");
            
            if(!list.isEmpty())
            {
                result.accumulate(list.get(0).intValue());
            }
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to get requested annotation count from repository because of a repository exception", ex);
        }
        catch(final QueryEvaluationException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to get requested annotation count from repository because of a query evaluation exception",
                    ex);
        }
        catch(final MalformedQueryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to get requested annotation count from repository because of a malformed query exception",
                    ex);
        }
        
        return result;
    }
    
    @Override
    public ResourceCount getAnnotationCountsByOntologyTermUri(final URI ontologyTermUri, final Ontology nextOntology,
            final boolean recursive)
    {
        final ResourceCount result = new UriCountImpl(0);
        final StringBuilder selectQuery = new StringBuilder(200);
        
        if(nextOntology != null
                && this.getVersionedOntology(nextOntology.getOntologyUri(), nextOntology.getOntologyVersionUri()) == null)
        {
            return result;
        }
        
        // If nextOntology is not defined || recursive == false
        if(nextOntology == null || recursive == false)
        {
            // Count all of the annotations directly referencing the given ontology term uri
            selectQuery.append(" SELECT (COUNT(DISTINCT ?annotationUri) AS ?countAnnotations) ");
            selectQuery.append(" WHERE { ");
            selectQuery.append(" GRAPH <");
            selectQuery.append(SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            selectQuery.append("> { ");
            selectQuery.append(" ?annotationUri <");
            selectQuery.append(RDF.TYPE.stringValue());
            selectQuery.append("> <");
            selectQuery.append(OasConstants.TAGGING_TAGGING);
            selectQuery.append("> . ");
            selectQuery.append(" }");
            selectQuery.append(" GRAPH ?annotationUri { ");
            selectQuery.append(" ?annotationUri <");
            selectQuery.append(OasConstants.MOAT_TAG_MEANING.stringValue());
            selectQuery.append("> <");
            selectQuery.append(ontologyTermUri.stringValue());
            selectQuery.append("> . ");
            selectQuery.append(" ?annotationUri <");
            selectQuery.append(OasConstants.TAGGING_TAGGED_RESOURCE.stringValue());
            selectQuery.append("> ?annotatedObjectUri . ");
            selectQuery.append(" } ");
            selectQuery.append(" }");
            
            try
            {
                final List<Literal> list =
                        this.executeSparqlSelectQueryOnSingleLiteralVariable(selectQuery.toString(), "countAnnotations");
                
                if(!list.isEmpty())
                {
                    result.accumulate(list.get(0).intValue());
                }
            }
            catch(final RepositoryException ex)
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                        "Failed to get requested annotation count from repository because of a repository exception",
                        ex);
            }
            catch(final QueryEvaluationException ex)
            {
                throw new ResourceException(
                        Status.SERVER_ERROR_INTERNAL,
                        "Failed to get requested annotation count from repository because of a query evaluation exception",
                        ex);
            }
            catch(final MalformedQueryException ex)
            {
                throw new ResourceException(
                        Status.SERVER_ERROR_INTERNAL,
                        "Failed to get requested annotation count from repository because of a malformed query exception",
                        ex);
            }
            
        }
        // If nextOntology is defined && recursive == true
        else if(nextOntology != null && recursive == true)
        {
            // Find all of the children for the given ontology term recursively, including the given
            // ontology term
            final Map<URI, Collection<URI>> children = this.getChildren(nextOntology, ontologyTermUri, true, 10000, 0);
            
            for(final URI nextChildUri : children.keySet())
            {
                // Count all of the annotations referencing the parent term
                final ResourceCount tempCount = this.getAnnotationCountsByOntologyTermUri(nextChildUri, null, false);
                
                result.accumulate(tempCount.getCount());
                
                for(final URI nextUri : children.get(nextChildUri))
                {
                    // Count all of the annotations referencing each child term
                    final ResourceCount tempChildrenCount =
                            this.getAnnotationCountsByOntologyTermUri(nextUri, null, false);
                    
                    result.accumulate(tempChildrenCount.getCount());
                }
            }
        }
        
        return result;
    }
    
    @Override
    public Collection<Annotation> getAnnotationsByAnnotatedObjectType(final URI annotatedObjectType)
    {
        final Collection<Annotation> results = new ArrayList<Annotation>();
        
        try
        {
            final StringBuilder selectQuery = new StringBuilder(200);
            
            // XXX: Should this count distinct annotations, or distinct annotated objects?
            
            selectQuery.append(" SELECT ?annotationUri WHERE { ");
            selectQuery.append(" GRAPH <");
            selectQuery.append(SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            selectQuery.append("> { ");
            selectQuery.append(" ?annotationUri <");
            selectQuery.append(RDF.TYPE.stringValue());
            selectQuery.append("> <");
            selectQuery.append(OasConstants.TAGGING_TAGGING);
            selectQuery.append("> . ");
            selectQuery.append(" }");
            selectQuery.append(" GRAPH ?annotationUri { ");
            selectQuery.append(" ?annotationUri <");
            selectQuery.append(OasConstants.TAGGING_TAGGED_RESOURCE.stringValue());
            selectQuery.append("> ?taggedResource . ");
            selectQuery.append(" ?taggedResource <");
            selectQuery.append(RDF.TYPE);
            selectQuery.append("> <");
            selectQuery.append(annotatedObjectType.stringValue());
            selectQuery.append("> . } }");
            
            for(final URI nextAnnotationUri : this.executeSparqlSelectQueryOnSingleUriVariable(selectQuery.toString(),
                    "annotationUri"))
            {
                final StringBuilder constructQuery = new StringBuilder(200);
                constructQuery.append(" CONSTRUCT { <");
                constructQuery.append(nextAnnotationUri.stringValue());
                constructQuery.append("> <");
                constructQuery.append(RDF.TYPE.stringValue());
                constructQuery.append("> <");
                constructQuery.append(OasConstants.TAGGING_TAGGING);
                constructQuery.append("> . ");
                constructQuery.append("<");
                constructQuery.append(nextAnnotationUri.stringValue());
                constructQuery.append("> <");
                constructQuery.append(OasConstants.TAGGING_TAGGED_RESOURCE.stringValue());
                constructQuery.append("> ?taggedResource . ");
                constructQuery.append(" ?taggedResource <");
                constructQuery.append(RDF.TYPE);
                constructQuery.append("> <");
                constructQuery.append(annotatedObjectType.stringValue());
                constructQuery.append("> . ");
                constructQuery.append(" } WHERE { ");
                
                constructQuery.append(" GRAPH <");
                constructQuery.append(nextAnnotationUri.stringValue());
                constructQuery.append("> { ");
                
                constructQuery.append(" <");
                constructQuery.append(nextAnnotationUri.stringValue());
                constructQuery.append("> <");
                constructQuery.append(RDF.TYPE.stringValue());
                constructQuery.append("> <");
                constructQuery.append(OasConstants.TAGGING_TAGGING);
                constructQuery.append("> . ");
                constructQuery.append(" <");
                constructQuery.append(nextAnnotationUri.stringValue());
                constructQuery.append("> <");
                constructQuery.append(OasConstants.TAGGING_TAGGED_RESOURCE.stringValue());
                constructQuery.append("> ?taggedResource . ");
                constructQuery.append(" ?taggedResource <");
                constructQuery.append(RDF.TYPE);
                constructQuery.append("> <");
                constructQuery.append(annotatedObjectType.stringValue());
                constructQuery.append("> . ");
                constructQuery.append(" } ");
                constructQuery.append(" } ");
                
                final Collection<Statement> queryResults = this.executeSparqlConstructQuery(constructQuery.toString());
                
                results.add(new AnnotationImpl(queryResults));
            }
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final MalformedQueryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final QueryEvaluationException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        
        return results;
    }
    
    @Override
    public Collection<Annotation> getAnnotationsByAnnotatedObjectUri(final URI annotatedObjectUri)
    {
        final Collection<Annotation> results = new ArrayList<Annotation>();
        
        try
        {
            final StringBuilder selectQuery = new StringBuilder();
            
            selectQuery.append(" SELECT ?annotationUri ");
            selectQuery.append(" WHERE { ");
            selectQuery.append(" GRAPH <");
            selectQuery.append(SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            selectQuery.append("> { ");
            selectQuery.append(" ?annotationUri <");
            selectQuery.append(RDF.TYPE.stringValue());
            selectQuery.append("> <");
            selectQuery.append(OasConstants.TAGGING_TAGGING);
            selectQuery.append("> . ");
            selectQuery.append(" }");
            selectQuery.append(" GRAPH ?annotationUri { ");
            selectQuery.append(" ?annotationUri <");
            selectQuery.append(OasConstants.TAGGING_TAGGED_RESOURCE.stringValue());
            selectQuery.append("> <");
            selectQuery.append(annotatedObjectUri.stringValue());
            selectQuery.append("> . ");
            selectQuery.append(" OPTIONAL { <");
            selectQuery.append(annotatedObjectUri.stringValue());
            selectQuery.append("> <");
            selectQuery.append(RDF.TYPE);
            selectQuery.append("> ?annotatedObjectType . } ");
            selectQuery.append(" } }");
            
            for(final URI nextAnnotationUri : this.executeSparqlSelectQueryOnSingleUriVariable(selectQuery.toString(),
                    "annotationUri"))
            {
                results.add(this.getAnnotation(nextAnnotationUri));
            }
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final MalformedQueryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final QueryEvaluationException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        
        return results;
    }
    
    @Override
    public Collection<Annotation> getAnnotationsByOntologyTermUri(final URI ontologyTermUri)
    {
        final Collection<Annotation> results = new ArrayList<Annotation>();
        
        try
        {
            final StringBuilder selectQuery = new StringBuilder("SELECT ");
            
            selectQuery.append("?annotationUri WHERE { ");
            selectQuery.append(" GRAPH <");
            selectQuery.append(SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            selectQuery.append("> { ");
            selectQuery.append(" ?annotationUri a <");
            selectQuery.append(OasConstants.TAGGING_TAGGING);
            selectQuery.append("> . ");
            selectQuery.append(" }");
            selectQuery.append(" GRAPH ?annotationUri { ");
            selectQuery.append(" ?annotationUri <");
            selectQuery.append(OasConstants.MOAT_TAG_MEANING.stringValue());
            selectQuery.append("> <");
            selectQuery.append(ontologyTermUri.stringValue());
            selectQuery.append("> . ?annotationUri <");
            selectQuery.append(OasConstants.TAGGING_TAGGED_RESOURCE.stringValue());
            selectQuery.append("> ?annotatedObjectUri . OPTIONAL { ?annotatedObjectUri a ?annotatedObjectType . } } }");
            
            for(final URI nextAnnotationUri : this.executeSparqlSelectQueryOnSingleUriVariable(selectQuery.toString(),
                    "annotationUri"))
            {
                results.add(this.getAnnotation(nextAnnotationUri));
            }
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final MalformedQueryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final QueryEvaluationException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        
        return results;
    }
    
    @Override
    public Map<URI, Collection<URI>> getChildren(final Ontology ontology, final URI parentUri, final boolean recursive,
            final int limit, final int offset)
    {
        final DatasetImpl queryDataset = new DatasetImpl();
        queryDataset.addDefaultGraph(ontology.getOntologyVersionUri());
        queryDataset.addDefaultGraph(ontology.getInferredOntologyUri());
        
        final StringBuilder queryString = new StringBuilder(256);
        
        // select all distinct parent/child combinations using the query that fits this ontologies
        // structure
        queryString.append("SELECT DISTINCT ?parent ?child WHERE { ");
        
        if(recursive)
        {
            queryString.append(ontology.getRecursiveChildQuery());
        }
        else
        {
            queryString.append(ontology.getDirectChildQuery());
        }
        
        queryString.append(" } ");
        
        // TODO: Test sorting along with the limits
        queryString.append(" ORDER BY DESC(?parent) DESC(?child) ");
        
        // TODO: Test these values
        queryString.append(" LIMIT ");
        queryString.append(limit);
        queryString.append(" OFFSET ");
        queryString.append(offset);
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            final TupleQuery graphQuery = conn.prepareTupleQuery(QueryLanguage.SPARQL, queryString.toString());
            
            graphQuery.setDataset(queryDataset);
            
            graphQuery.clearBindings();
            
            if(parentUri == null)
            {
                graphQuery.setBinding("parent", OasConstants.OWL_THING);
            }
            else
            {
                graphQuery.setBinding("parent", parentUri);
            }
            
            final ConcurrentMap<URI, Collection<URI>> results = new ConcurrentHashMap<>();
            
            final TupleQueryResult queryResult = graphQuery.evaluate();
            
            while(queryResult.hasNext())
            {
                final BindingSet next = queryResult.next();
                
                final URI nextParentUri;
                
                // HACK: FIXME:
                if(next.hasBinding("parent"))
                {
                    nextParentUri = (URI)next.getBinding("parent").getValue();
                }
                else if(parentUri != null)
                {
                    nextParentUri = parentUri;
                }
                else
                {
                    nextParentUri = OasConstants.OWL_THING;
                }
                
                final URI nextChildUri = (URI)next.getBinding("child").getValue();
                
                final ArrayList<URI> newList = new ArrayList<URI>(2);
                newList.add(nextChildUri);
                
                // optimistically create a new collection and use the ConcurrentMap API to fall
                // through to adding to an existing list if one was created prior to this point
                final Collection<URI> putIfAbsent = results.putIfAbsent(nextParentUri, newList);
                
                if(putIfAbsent != null)
                {
                    putIfAbsent.add(nextChildUri);
                }
            }
            
            return results;
        }
        catch(RepositoryException | MalformedQueryException | QueryEvaluationException e)
        {
            throw new RuntimeException(e);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException e)
                {
                    this.log.error("Found unexpected exception while closing repository connection", e);
                }
            }
        }
    }
    
    @Override
    public ConcurrentMap<URI, ResourceCount> getChildrenCount(final Ontology ontology, final List<URI> parentUri,
            final boolean recursive)
    {
        final DatasetImpl queryDataset = new DatasetImpl();
        queryDataset.addDefaultGraph(ontology.getOntologyVersionUri());
        queryDataset.addDefaultGraph(ontology.getInferredOntologyUri());
        
        final StringBuilder queryString = new StringBuilder(256);
        
        queryString.append("SELECT DISTINCT ?parent (COUNT(DISTINCT ?child) as ?childCount)  WHERE { ");
        
        if(recursive)
        {
            queryString.append(ontology.getRecursiveChildQuery());
        }
        else
        {
            queryString.append(ontology.getDirectChildQuery());
        }
        
        queryString.append(" } ");
        
        // this statement is necessary to aggregate the children to get a result
        queryString.append(" GROUP BY ?parent ");
        
        queryString.append(" VALUES (?parent) { ");
        if(parentUri == null || parentUri.isEmpty())
        {
            queryString.append(" ( <");
            queryString.append(OWL.THING.stringValue());
            queryString.append("> )");
        }
        else
        {
            for(URI nextParentURI : parentUri)
            {
                queryString.append(" ( <");
                queryString.append(nextParentURI.stringValue());
                queryString.append("> ) ");
            }
        }
        queryString.append(" } ");
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            // this.log.info("query: {}", queryString);
            
            final TupleQuery graphQuery = conn.prepareTupleQuery(QueryLanguage.SPARQL, queryString.toString());
            
            graphQuery.setDataset(queryDataset);
            
            final TupleQueryResult queryResult = graphQuery.evaluate();
            
            ConcurrentMap<URI, ResourceCount> resultMap = new ConcurrentHashMap<URI, ResourceCount>();
            
            while(queryResult.hasNext())
            {
                final BindingSet next = queryResult.next();
                
                if(next.hasBinding("parent"))
                {
                    Value nextParentValue = next.getBinding("parent").getValue();
                    if(nextParentValue instanceof URI)
                    {
                        if(next.hasBinding("childCount"))
                        {
                            final Literal childCount = (Literal)next.getBinding("childCount").getValue();
                            
                            // Optimistically put a new count into the map using the ConcurrentMap
                            // API We use ConcurrentMap API to cope with the possibility that Sesame
                            // may again in the future return multiple bindings for a count for a
                            // particular URI
                            ResourceCount putIfAbsent =
                                    resultMap
                                            .putIfAbsent((URI)nextParentValue, new UriCountImpl(childCount.intValue()));
                            if(putIfAbsent != null)
                            {
                                putIfAbsent.accumulate(childCount.intValue());
                            }
                        }
                        else
                        {
                            this.log.warn("childCount binding was not set for a results set: bindingSet={}", next);
                        }
                    }
                    else
                    {
                        this.log.warn("Found matching parent resource that was not a URI: bindingSet={}", next);
                    }
                }
                
            }
            return resultMap;
        }
        catch(RepositoryException | MalformedQueryException | QueryEvaluationException e)
        {
            throw new RuntimeException(e);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException e)
                {
                    this.log.error("Found unexpected exception while closing repository connection", e);
                }
            }
        }
    }
    
    @Override
    public List<Literal> getLabelsForUri(final Ontology nextOntology, final URI nextChildUri)
    {
        RepositoryConnection conn = null;
        
        // this.addLogDetailsForRequest(annotationIdUri.stringValue(), request,
        // DataHandlerConstants.DELETE);
        
        try
        {
            conn = this.repository.getConnection();
            
            URI[] contexts;
            
            if(nextOntology != null)
            {
                contexts = nextOntology.getContextUris();
            }
            else
            {
                List<URI> allCurrentVersionContexts = new ArrayList<URI>();
                
                List<Ontology> ontologies = this.getOntologies(0, 1000);
                
                if(ontologies.size() == 0)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_NOT_FOUND, "Could not find label");
                }
                
                for(Ontology nextCurrentOntology : ontologies)
                {
                    allCurrentVersionContexts.addAll(Arrays.asList(nextCurrentOntology.getContextUris()));
                }
                
                if(allCurrentVersionContexts.size() == 0)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_NOT_FOUND, "Could not find label");
                }
                
                contexts = allCurrentVersionContexts.toArray(new URI[allCurrentVersionContexts.size()]);
            }
            
            final List<Literal> results = new ArrayList<Literal>(OasConstants.PREFERRED_LABEL_PROPERTIES.size());
            
            for(final URI nextLabelUri : OasConstants.PREFERRED_LABEL_PROPERTIES)
            {
                final List<Statement> nextList =
                        Iterations.asList(conn.getStatements(nextChildUri, nextLabelUri, null, true, contexts));
                
                for(final Statement nextStatement : nextList)
                {
                    if(nextStatement.getObject() instanceof Literal)
                    {
                        results.add((Literal)nextStatement.getObject());
                    }
                }
            }
            
            return results;
        }
        catch(final RepositoryException e)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Could not find label for URI.");
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to rollback the transaction", ex);
            }
        }
    }
    
    @Override
    public Ontology setLatestOntology(final URI ontologyUri, final URI versionUri)
    {
        SesameUtils.setLatestVersion(ontologyUri, versionUri, repository, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
        
        return getLatestOntology(ontologyUri);
    }
    
    @Override
    public Ontology getLatestOntology(final URI ontologyUri)
    {
        Ontology result = null;
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            final List<Statement> versionList =
                    Iterations.asList(conn.getStatements(ontologyUri, OasConstants.OMV_CURRENTVERSION, null, false,
                            SesameUtils.ONTOLOGY_MANAGER_GRAPH));
            
            if(versionList.size() > 1)
            {
                this.log.error("Current version marker for ontology was duplicated ontologyUri={}", ontologyUri);
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                        "Current version marker for ontology was duplicated ontologyUri=" + ontologyUri.stringValue());
            }
            
            if(versionList.size() == 1)
            {
                result = this.getVersionedOntology(ontologyUri, (URI)versionList.get(0).getObject());
            }
            
            if(result == null)
            {
                final List<Statement> ontologyList =
                        Iterations.asList(conn.getStatements(null, OasConstants.OWL_VERSIONIRI, ontologyUri, false,
                                SesameUtils.ONTOLOGY_MANAGER_GRAPH));
                
                if(ontologyList.size() > 1)
                {
                    this.log.error(
                            "Version IRI for ontology was duplicated across different ontologies ontologyUri={} ontologyList={}",
                            ontologyUri, ontologyList);
                    throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                            "Version IRI for ontology was duplicated across different ontologies ontologyUri="
                                    + ontologyUri.stringValue());
                }
                
                if(ontologyList.size() == 1)
                {
                    result = this.getVersionedOntology((URI)ontologyList.get(0).getSubject(), ontologyUri);
                }
            }
        }
        catch(final RepositoryException rex)
        {
            this.log.error("Repository access was not possible", rex);
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Repository access was not possible", rex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
        
        return result;
    }
    
    @Override
    public Model getLogDetails(final int offset, final int limit)
    {
        final Model model = new LinkedHashModel();
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            // DebugUtils.printContexts(conn);
            // DebugUtils.printContents(conn, SesameUtils.AUDIT_LOG_GRAPH);
            
            final ValueFactory vf = conn.getValueFactory();
            
            final List<URI> recordList =
                    this.executeSparqlSelectQueryOnSingleUriVariable("SELECT DISTINCT ?logRecord WHERE { GRAPH<"
                            + SesameUtils.AUDIT_LOG_GRAPH.stringValue() + "> { ?logRecord a <"
                            + OasConstants.URN_TAG_AUDITLOG_RECORD.stringValue() + "> . ?logRecord <"
                            + OasConstants.DC_TERMS_MODIFIED.stringValue()
                            + "> ?timestamp . } } ORDER BY DESC(?timestamp) OFFSET " + offset + " LIMIT " + limit,
                            "logRecord");
            
            for(final URI nextRecord : recordList)
            {
                Iterations.addAll(conn.getStatements(nextRecord, null, null, true, SesameUtils.AUDIT_LOG_GRAPH), model);
            }
        }
        catch(RepositoryException | MalformedQueryException | QueryEvaluationException re)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to get audit log, as an internal exception occurred.", re);
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                this.log.error("ERROR: A RepositoryException occured while trying to rollback the transaction", ex);
            }
        }
        
        return model;
    }
    
    /*
     * Fetches ontologies, sorting by the ontology URI for paging purposes.
     */
    @Override
    public List<Ontology> getOntologies(final int offset, final int limit)
    {
        final List<Ontology> results = new ArrayList<Ontology>();
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            if(conn.hasStatement(null, RDF.TYPE, OWL.ONTOLOGY, false, SesameUtils.ONTOLOGY_MANAGER_GRAPH))
            {
                final String query =
                        "SELECT ?ontologyUri FROM <" + SesameUtils.ONTOLOGY_MANAGER_GRAPH
                                + "> WHERE { ?ontologyUri a <" + OWL.ONTOLOGY.stringValue()
                                + "> } ORDER BY ?ontologyUri OFFSET " + offset + " LIMIT " + limit;
                final Collection<URI> ontologyUris =
                        this.executeSparqlSelectQueryOnSingleUriVariable(query, "ontologyUri");
                
                for(final URI nextOntologyUri : ontologyUris)
                {
                    final Ontology ontology = this.getLatestOntology(nextOntologyUri);
                    
                    if(ontology != null)
                    {
                        results.add(ontology);
                    }
                }
            }
        }
        catch(final RepositoryException rex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Repository access was not possible", rex);
        }
        catch(final MalformedQueryException mqe)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Query was not formed correctly", mqe);
        }
        catch(final QueryEvaluationException qee)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Query did not evaluate correctly", qee);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
        
        return results;
    }
    
    @Override
    public List<Ontology> getOntologiesByUri(final URI ontologyUri) throws ResourceException
    {
        final List<Ontology> results = new ArrayList<Ontology>();
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            if(conn.hasStatement(ontologyUri, RDF.TYPE, OWL.ONTOLOGY, false, SesameUtils.ONTOLOGY_MANAGER_GRAPH))
            {
                RepositoryResult<Statement> versionList = null;
                
                try
                {
                    versionList =
                            conn.getStatements(ontologyUri, OasConstants.OWL_VERSIONIRI, null, false,
                                    SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    
                    while(versionList.hasNext())
                    {
                        final Statement nextStatement = versionList.next();
                        
                        if(this.log.isTraceEnabled())
                        {
                            this.log.trace("nextManagedOntologyVersion={}", nextStatement.getObject().stringValue());
                        }
                        
                        if(nextStatement.getObject() instanceof URI)
                        {
                            if(this.log.isTraceEnabled())
                            {
                                this.log.trace("nextManagedOntologyVersion instanceof URI");
                            }
                            
                            final Ontology managedOntology =
                                    this.getVersionedOntology(ontologyUri, (URI)nextStatement.getObject());
                            
                            if(managedOntology != null)
                            {
                                if(this.log.isTraceEnabled())
                                {
                                    this.log.trace("managedOntology was not null");
                                }
                                results.add(managedOntology);
                            }
                            else
                            {
                                this.log.debug("Found null managedOntology ontologyUri= {} nextVersion={}",
                                        ontologyUri.stringValue(), nextStatement.getObject());
                            }
                        }
                        else
                        {
                            this.log.warn("Found non-URI for OWL:VERSIONIRI value={}", nextStatement.getObject());
                        }
                    }
                }
                finally
                {
                    if(versionList != null)
                    {
                        versionList.close();
                    }
                }
            }
        }
        catch(final RepositoryException rex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Repository access was not possible", rex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
        
        return results;
    }
    
    @Override
    public Ontology getVersionedOntology(final URI ontologyUri, final URI managedOntologyVersionUri)
    {
        if(ontologyUri == null || managedOntologyVersionUri == null)
        {
            throw new IllegalArgumentException("getManagedOntology cannot be called with null parameters");
        }
        
        Ontology result = null;
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            if(conn.hasStatement(ontologyUri, RDF.TYPE, OWL.ONTOLOGY, false, SesameUtils.ONTOLOGY_MANAGER_GRAPH))
            {
                if(conn.hasStatement(ontologyUri, OasConstants.OWL_VERSIONIRI, managedOntologyVersionUri, false,
                        SesameUtils.ONTOLOGY_MANAGER_GRAPH))
                {
                    result =
                            SesameUtils.createOntologyFromRepository(conn, ontologyUri, managedOntologyVersionUri,
                                    SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                }
                else
                {
                    this.log.debug(
                            "Could not find ontology version in ontology manager, returning null: ontologyUri={} managedOntologyVersionUri={}",
                            ontologyUri, managedOntologyVersionUri);
                }
            }
            else
            {
                this.log.debug("Could not find ontology in ontology manager, returning null: ontologyUri={}",
                        ontologyUri);
            }
        }
        catch(final RepositoryException rex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Repository access was not possible", rex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    this.log.error("Found repository exception while closing connection", ex);
                }
            }
        }
        
        return result;
    }
    
    @Override
    public Collection<Annotation> listAnnotations(final int offset, final int limit)
    {
        final Collection<Annotation> allResults = new ArrayList<Annotation>();
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = this.repository.getConnection();
            
            final RepositoryResult<Statement> statements =
                    conn.getStatements(null, RDF.TYPE, OasConstants.TAGGING_TAGGING, true,
                            SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            
            while(statements.hasNext())
            {
                final Statement nextResult = statements.next();
                
                final Annotation annotation = this.getAnnotation((URI)nextResult.getSubject());
                
                if(annotation != null)
                {
                    allResults.add(annotation);
                }
            }
        }
        catch(final RepositoryException ex)
        {
            this.log.error("Failed to complete search, as an exception occurred.", ex);
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete search, as an exception occurred.", ex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException e)
                {
                    this.log.error("Found repository exception while closing connection", e);
                }
            }
        }
        
        return allResults;
    }
    
    @Override
    public List<SearchResult> searchAllOntologies(final String searchTerm, final int limit, final int offset)
    {
        final Collection<Ontology> ontologies = this.getOntologies(0, 1000);
        
        return this.searchOntologyTermsInternal(ontologies, searchTerm, limit, offset);
    }
    
    @Override
    public Collection<Annotation> searchAnnotations(final String searchTerm, final int limit, final int offset)
    {
        if(searchTerm == null)
        {
            throw new IllegalArgumentException("Cannot search for annotations without a search term");
        }
        final StringBuilder sb = new StringBuilder();
        
        sb.append("SELECT ?uri WHERE {");
        sb.append(" GRAPH <");
        sb.append(SesameUtils.ANNOTATION_MANAGEMENT_GRAPH.stringValue());
        sb.append("> { ");
        sb.append(" ?uri a <");
        sb.append(OasConstants.TAGGING_TAGGING.stringValue());
        sb.append("> .");
        sb.append(" } ");
        if(!searchTerm.trim().isEmpty())
        {
            sb.append(" GRAPH ?uri { ");
            sb.append(" ?uri a <");
            sb.append(OasConstants.TAGGING_TAGGING.stringValue());
            sb.append("> . ");
            sb.append(" ?uri ?predicate ?literal . ");
            sb.append(" FILTER(CONTAINS(?literal, \"");
            sb.append(NTriplesUtil.escapeString(searchTerm));
            sb.append("\"))");
            sb.append(" } ");
        }
        
        sb.append(" } ");
        sb.append(" ORDER BY ?uri ");// ?literal ");
        sb.append(" LIMIT ");
        sb.append(limit);
        sb.append(" OFFSET ");
        sb.append(offset);
        
        final Collection<Annotation> allResults = new ArrayList<>();
        
        RepositoryConnection conn = null;
        
        // TODO: Execute query with permissions of current authenticated user
        
        try
        {
            conn = this.repository.getConnection();
            
            final TupleQuery query = conn.prepareTupleQuery(QueryLanguage.SPARQL, sb.toString());
            // DatasetImpl dataset = new DatasetImpl();
            // dataset.addDefaultGraph(SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            // query.setDataset(dataset);
            
            final TupleQueryResult result = query.evaluate();
            
            while(result.hasNext())
            {
                final BindingSet nextResult = result.next();
                
                final Annotation annotation = this.getAnnotation((URI)nextResult.getBinding("uri").getValue());
                if(annotation != null)
                {
                    allResults.add(annotation);
                }
            }
        }
        catch(final RepositoryException | MalformedQueryException | QueryEvaluationException ex)
        {
            this.log.error("Failed to complete search, as an exception occurred.", ex);
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete search, as an exception occurred.", ex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException e)
                {
                    this.log.error("Found repository exception while closing connection", e);
                }
            }
        }
        
        return allResults;
    }
    
    @Override
    public List<SearchResult> searchOntologies(final Collection<Ontology> ontologies, final String searchTerm,
            final int limit, final int offset)
    {
        final List<SearchResult> results = new ArrayList<SearchResult>();
        
        results.addAll(this.searchOntologyTermsInternal(ontologies, searchTerm, limit, offset));
        
        return results;
    }
    
    private List<SearchResult> searchOntologyLabelInternal(final Ontology nextOntology, final String searchTerm,
            final int limit, final int offset)
    {
        final StringBuilder sb = new StringBuilder();
        
        sb.append("SELECT DISTINCT ?uri ?literal WHERE { ");
        sb.append(" GRAPH ?ontology { ");
        sb.append("?uri a <");
        sb.append(OWL.ONTOLOGY.stringValue());
        sb.append("> . ");
        sb.append(" ?uri ?property ?literal . ");
        sb.append(" FILTER(CONTAINS(?literal, \"");
        sb.append(NTriplesUtil.escapeString(searchTerm));
        sb.append("\")) } } ");
        sb.append(" ORDER BY ?uri ?literal");
        sb.append(" LIMIT ");
        sb.append(limit);
        sb.append(" OFFSET ");
        sb.append(offset);
        
        final List<SearchResult> allResults = new LinkedList<>();
        
        RepositoryConnection conn = null;
        
        // TODO: Execute query with permissions of current authenticated user
        
        try
        {
            conn = this.repository.getConnection();
            
            final TupleQuery query = conn.prepareTupleQuery(QueryLanguage.SPARQL, sb.toString());
            query.setBinding("ontology", nextOntology.getOntologyVersionUri());
            
            final TupleQueryResult result = query.evaluate();
            
            final Set<BindingSet> initialResults = Iterations.asSet(result);
            
            final Map<URI, Value> matchingUris = new HashMap<URI, Value>();
            
            for(final BindingSet nextResult : initialResults)
            {
                if(nextResult.getBinding("uri").getValue() instanceof URI)
                {
                    if(!matchingUris.containsKey(nextResult.getBinding("uri").getValue()))
                    {
                        matchingUris.put((URI)nextResult.getBinding("uri").getValue(), nextResult.getBinding("literal")
                                .getValue());
                    }
                    else
                    {
                        this.log.info("Found duplicate ontology search result: {}", nextResult);
                    }
                }
                else
                {
                    // No way of communicating search results for ontologies who link literals to
                    // blank
                    // nodes. Linked Data is the only way to communicate these things successfully
                    this.log.warn("Found a non-URI for label search: {}", nextResult);
                }
            }
            
            for(final URI nextResultUri : matchingUris.keySet())
            {
                final String labelsQueryString =
                        "SELECT DISTINCT ?rdfslabel WHERE { GRAPH ?ontology { ?uri <http://www.w3.org/2000/01/rdf-schema#label> ?rdfslabel . } }";
                
                final TupleQuery labelsQuery = conn.prepareTupleQuery(QueryLanguage.SPARQL, labelsQueryString);
                labelsQuery.setBinding("ontology", nextOntology.getOntologyVersionUri());
                labelsQuery.setBinding("uri", nextResultUri);
                
                final TupleQueryResult labelsResult = labelsQuery.evaluate();
                
                final List<String> labels = new ArrayList<String>();
                
                // FIXME: Rewrite to favour rdfs:label
                for(final BindingSet test : Iterations.asList(labelsResult))
                {
                    if(test.hasBinding("rdfslabel"))
                    {
                        labels.add(test.getBinding("rdfslabel").getValue().stringValue());
                    }
                }
                
                if(labels.isEmpty())
                {
                    // If there were no rdfs:label's available, then resort to using the literal we
                    // initially found
                    labels.add(matchingUris.get(nextResultUri).stringValue());
                }
                
                allResults.add(new SearchResultImpl(nextOntology, nextResultUri, labels));
            }
        }
        catch(final RepositoryException | MalformedQueryException | QueryEvaluationException ex)
        {
            this.log.error("Failed to complete search, as an exception occurred.", ex);
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete search, as an exception occurred.", ex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException e)
                {
                    this.log.error("Found repository exception while closing connection", e);
                }
            }
        }
        
        return allResults;
    }
    
    @Override
    public List<SearchResult> searchOntologyLabels(final String searchTerm, final int limit, final int offset)
    {
        final Collection<Ontology> ontologies = this.getOntologies(0, 1000);
        
        final List<SearchResult> results = new ArrayList<SearchResult>();
        
        for(final Ontology nextOntology : ontologies)
        {
            if(nextOntology == null)
            {
                continue;
            }
            results.addAll(this.searchOntologyLabelInternal(nextOntology, searchTerm, limit, offset));
        }
        return results;
    }
    
    private List<SearchResult> searchOntologyTermsInternal(final Collection<Ontology> ontologies,
            final String searchTerm, final int limit, final int offset)
    {
        final String sparqlQuery =
                "SELECT DISTINCT ?ontology ?uri WHERE { GRAPH ?ontology { ?uri ?predicate ?literal . FILTER(CONTAINS(?literal, \""
                        + NTriplesUtil.escapeString(searchTerm) + "\")) } } LIMIT " + limit + " OFFSET " + offset;
        
        final List<SearchResult> allResults = new LinkedList<>();
        
        RepositoryConnection conn = null;
        
        // TODO: Execute query with permissions of current authenticated user
        
        try
        {
            conn = this.repository.getConnection();
            
            final TupleQuery query = conn.prepareTupleQuery(QueryLanguage.SPARQL, sparqlQuery);
            
            // Use a single model to accumulate the labels
            final Model labelsModel = new LinkedHashModel();
            
            for(final Ontology nextOntology : ontologies)
            {
                if(nextOntology == null)
                {
                    continue;
                }
                
                query.clearBindings();
                query.setBinding("ontology", nextOntology.getOntologyVersionUri());
                final TupleQueryResult result = query.evaluate();
                
                while(result.hasNext())
                {
                    final BindingSet nextResult = result.next();
                    
                    // No way of communicating search results for ontologies who link literals to
                    // blank
                    // nodes. Linked Data is the only way to communicate these things successfully
                    if(nextResult.getBinding("uri").getValue() instanceof URI)
                    {
                        // Reset the model, to save recreating it each time
                        labelsModel.clear();
                        
                        final URI nextResultUri = (URI)nextResult.getBinding("uri").getValue();
                        
                        conn.exportStatements(nextResultUri, RDFS.LABEL, null, true,
                                new StatementCollector(labelsModel), nextOntology.getOntologyVersionUri());
                        
                        final List<String> labels = new ArrayList<String>();
                        
                        for(final Value test : labelsModel.objects())
                        {
                            if(test instanceof Literal)
                            {
                                labels.add(((Literal)test).getLabel());
                            }
                        }
                        
                        if(!labels.isEmpty())
                        {
                            allResults.add(new SearchResultImpl(nextOntology, nextResultUri, labels));
                        }
                        else
                        {
                            allResults.add(new SearchResultImpl(nextOntology, nextResultUri));
                        }
                    }
                }
            }
        }
        catch(final RepositoryException | MalformedQueryException | QueryEvaluationException | RDFHandlerException ex)
        {
            this.log.error("Failed to complete search, as an exception occurred.", ex);
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete search, as an exception occurred.", ex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException e)
                {
                    this.log.error("Found repository exception while closing connection", e);
                }
            }
        }
        
        return allResults;
    }
    
    /**
     * Stores the annotation in the sesame repository using the annotationId as the graph/context
     * Also stores metadata such as when the tag was modifed (creation is modelled as modified) and
     * the IP address of the creator in a management graph.
     * 
     * @param nextAnnotation
     *            The annotation to store.
     */
    @Override
    public boolean storeAnnotation(final Annotation nextAnnotation)
    {
        // this.addLogDetailsForRequest(nextAnnotation.getAnnotationUniqueId().stringValue(),
        // request, DataHandlerConstants.CREATE);
        
        // RepositoryConnection conn = null;
        
        try
        {
            // conn = this.repository.getConnection();
            
            // conn.setAutoCommit(false);
            
            SesameUtils.toRDFRepository(nextAnnotation, this.repository, SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            
            // conn.add(SesameUtils
            // .toRDFRepository(nextAnnotation, null)
            // .getConnection()
            // .getStatements((Resource)null, (URI)null, (Value)null, true,
            // nextAnnotation.getAnnotationUniqueId()));
            //
            // If everything went as planned, we can commit the result
            // conn.commit();
            
            return true;
        }
        catch(final RepositoryException re)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete storage, as a repository exception occurred.", re);
        }
    }
    
    @Override
    public Ontology storeOntology(final Ontology nextOntology, final InputStream inputStream, final String baseUri,
            final String mimeType, final boolean deleteOldVersions) throws ResourceException
    {
        if(inputStream == null)
        {
            throw new IllegalArgumentException("Input Stream must not be null");
        }
        
        if(nextOntology == null)
        {
            throw new IllegalArgumentException("Ontology must not be null");
        }
        
        if(nextOntology.getOntologyUri() == null)
        {
            throw new IllegalArgumentException("Could not store ontology without an ontology URI");
        }
        
        String actualBaseUri = baseUri;
        
        if(baseUri == null)
        {
            actualBaseUri = nextOntology.getOntologyUri().stringValue();
        }
        
        if(mimeType == null)
        {
            throw new IllegalArgumentException("MIME-Type must not be null");
        }
        
        // Rely on the mime-type to fetch the parser factory, without going through the check-all
        // routine that ParsableOWLOntologyFactory goes through in its random walk through the
        // parsers to try to find the first one that doesn't fail
        
        final OWLOntologyFormatFactory formatFactory =
                OWLOntologyFormatFactoryRegistry.getInstance().getByMIMEType(mimeType);
        
        if(formatFactory == null)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Could not find an OWL compatible format matching the given mime type mimeType=" + mimeType);
        }
        
        if(formatFactory.isTextual())
        {
            // TODO: Is it dangerous to add UTF-8 using new InputStreamReader(inputStream,
            // StandardCharsets.UTF_8)
            final ReaderDocumentSource documentSource =
                    new ReaderDocumentSource(new BufferedReader(new InputStreamReader(inputStream,
                            StandardCharsets.UTF_8)), IRI.create(actualBaseUri), formatFactory);
            
            return this.storeOntologyInternal(nextOntology, documentSource, deleteOldVersions);
        }
        else
        {
            // FIXME: This is broken, but will be necessary to support Binary RDF
            final StreamDocumentSource documentSource =
                    new StreamDocumentSource(inputStream, IRI.create(actualBaseUri), formatFactory);
            
            return this.storeOntologyInternal(nextOntology, documentSource, deleteOldVersions);
        }
    }
    
    @Override
    public Ontology storeOntology(final Ontology nextOntology, final Reader inputReader, final String baseUri,
            final String mimeType, final boolean deleteOldVersions) throws ResourceException
    {
        if(inputReader == null)
        {
            throw new IllegalArgumentException("Reader must not be null");
        }
        
        if(nextOntology == null)
        {
            throw new IllegalArgumentException("Ontology must not be null");
        }
        
        String actualBaseUri = baseUri;
        
        if(baseUri == null)
        {
            actualBaseUri = nextOntology.getOntologyUri().stringValue();
        }
        
        if(nextOntology.getOntologyUri() == null)
        {
            throw new IllegalArgumentException("Could not store ontology without an ontology URI");
        }
        
        if(mimeType == null)
        {
            throw new IllegalArgumentException("MIME-Type must not be null");
        }
        
        // Rely on the mime-type to fetch the parser factory, without going through the check-all
        // routine that ParsableOWLOntologyFactory goes through in its random walk through the
        // parsers to try to find the first one that doesn't fail
        
        final OWLOntologyFormatFactory formatFactory =
                OWLOntologyFormatFactoryRegistry.getInstance().getByMIMEType(mimeType);
        
        if(formatFactory == null)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Could not find an OWL compatible format matching the given mime type mimeType=" + mimeType);
        }
        
        final ReaderDocumentSource documentSource =
                new ReaderDocumentSource(inputReader, IRI.create(baseUri), formatFactory);
        
        return this.storeOntologyInternal(nextOntology, documentSource, deleteOldVersions);
    }
    
    private Ontology storeOntologyInternal(final Ontology nextOntology, final OWLOntologyDocumentSource documentSource,
            final boolean deleteOldVersions) throws ResourceException
    {
        final Model concreteModel = new LinkedHashModel();
        final Model inferredModel = new LinkedHashModel();
        
        this.storeOntologyInternalOwl(nextOntology, documentSource, concreteModel, inferredModel);
        
        this.storeOntologyInternalSesame(nextOntology, deleteOldVersions, concreteModel, inferredModel);
        
        return nextOntology;
    }
    
    /**
     * Helper method to aggregate the OWLAPI functionality required by storeOntology.
     * 
     * @param nextOntology
     * @param documentSource
     * @param concreteModel
     * @param inferredModel
     * @throws ResourceException
     * @throws ReasonerInterruptedException
     * @throws TimeOutException
     * @throws InconsistentOntologyException
     * @throws OWLOntologyChangeException
     */
    protected void storeOntologyInternalOwl(final Ontology nextOntology,
            final OWLOntologyDocumentSource documentSource, final Model concreteModel, final Model inferredModel)
        throws ResourceException, ReasonerInterruptedException, TimeOutException, InconsistentOntologyException,
        OWLOntologyChangeException
    {
        // NOTE: This means that any supported non-RDF formats can be used for input, although
        // the export to RDF may lose some features, as some OWL features do not have serialisations
        // in terms of RDF triples
        final OWLParserFactory parserFactory =
                OWLParserFactoryRegistry.getInstance().getParserFactory(documentSource.getFormatFactory());
        
        if(parserFactory == null)
        {
            throw new ResourceException(Status.SERVER_ERROR_NOT_IMPLEMENTED,
                    "The given MIME-Type is not currently recognised: "
                            + documentSource.getFormatFactory().getDefaultMIMEType());
        }
        
        final List<OWLProfile> allProfiles = OWLProfileRegistry.getInstance().getProfiles();
        final List<OWLProfileReport> inProfileReports = new ArrayList<OWLProfileReport>(allProfiles.size());
        final List<OWLProfile> inProfiles = new ArrayList<OWLProfile>(allProfiles.size());
        final List<OWLProfileReport> outProfileReports = new ArrayList<OWLProfileReport>(allProfiles.size());
        final List<OWLProfile> outProfiles = new ArrayList<OWLProfile>(allProfiles.size());
        OWLOntology parsedOntology = null;
        OWLOntologyFormat parsedFormat = null;
        
        synchronized(this.manager)
        {
            final OWLParser parser = parserFactory.createParser(this.manager);
            
            try
            {
                parsedOntology = this.manager.getOntology(nextOntology.toOWLOntologyID());
            }
            catch(final RuntimeException re)
            {
                if(this.log.isDebugEnabled())
                {
                    this.log.debug("Starting OWLOntologyManagerImpl hack after getOntology");
                }
                // HUGE HACK: FIXME: TODO:
                // catch my hack that makes getOntology compatible with createOntology in cases
                // where the ontology was not stored in the manager, but the documentIRI mapping has
                // been stored
                // When this design choice is fixed by OWLAPI we may be able to remove this section
                try
                {
                    nextOntology.setOntologyVersionUri(SesameUtils.generateNewOntologyVersionUri(
                            nextOntology.getOntologyUri(), null, this.repository, SesameUtils.ONTOLOGY_MANAGER_GRAPH));
                    
                    parsedOntology = this.manager.createOntology(nextOntology.toOWLOntologyID());
                }
                catch(RepositoryException | OWLOntologyCreationException e)
                {
                    this.log.error("Failed to generate an ontology version Uri that was not in the manager previously",
                            e);
                    // throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    // "Failed to complete storage, as an exception occurred.", e);
                }
                if(this.log.isDebugEnabled())
                {
                    this.log.debug("Ending OWLOntologyManagerImpl hack after getOntology");
                }
            }
            
            // if the only did not already exist, then getOntology will return null and we need to
            // create a new ontology
            if(parsedOntology == null)
            {
                if(this.log.isDebugEnabled())
                {
                    this.log.debug("Ontology was not recognised, creating an ontology in the OWLOntology instance with the given ontologyIRI and ontologyVersionIRI");
                }
                try
                {
                    parsedOntology = this.manager.createOntology(nextOntology.toOWLOntologyID());
                }
                catch(final OWLOntologyCreationException e)
                {
                    throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                            "OWLOntologyManager could not create the Ontology", e);
                }
            }
            
            try
            {
                parsedFormat = parser.parse(documentSource, parsedOntology);
                
                if(!parsedFormat.getKey().equals(documentSource.getFormatFactory().getKey()))
                {
                    this.log.warn("Parser generated a different OWLOntologyFormat to what we were expecting expected: "
                            + documentSource.getFormatFactory().getKey() + " actual: " + parsedFormat.getKey());
                }
                
                if(parsedFormat instanceof RDFOntologyFormat)
                {
                    final RDFOntologyFormat parsedRdfFormat = (RDFOntologyFormat)parsedFormat;
                    
                    if(!parsedRdfFormat.getErrors().isEmpty())
                    {
                        this.log.warn("Found errors importing ontology into owlapi: {}", parsedRdfFormat.getErrors());
                    }
                }
                else if(this.log.isDebugEnabled())
                {
                    this.log.debug("Imported a non-RDF ontology: parsedFormat={}", parsedFormat);
                }
            }
            catch(OWLOntologyChangeException | UnloadableImportException | OWLParserException | IOException e)
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL, e);
            }
            for(final OWLProfile profile : allProfiles)
            {
                final OWLProfileReport report = profile.checkOntology(parsedOntology);
                
                if(report.isInProfile())
                {
                    inProfileReports.add(report);
                    inProfiles.add(profile);
                }
                else
                {
                    outProfileReports.add(report);
                    outProfiles.add(profile);
                }
            }
            
            // add the list of profiles to the ontology
            nextOntology.setProfiles(inProfiles);
            
            if(outProfileReports.isEmpty())
            {
                if(this.log.isDebugEnabled())
                {
                    this.log.debug("No errors found checking for ontology profiles");
                }
            }
            else
            {
                if(this.log.isDebugEnabled())
                {
                    this.log.debug("Found errors checking for ontology profiles: {}", outProfiles);
                    
                    for(final OWLProfileReport nextReport : outProfileReports)
                    {
                        this.log.debug("Found errors checking for ontology profile: profile={}",
                                nextReport.getProfile());
                    }
                }
            }
            
            final String reasonerName = OasConstants.OAS_REASONER;
            
            // TODO: Make this configurable
            final IRI reasonerProfile = OWLProfile.OWL2_DL;
            boolean inReasonerProfile = false;
            OWLProfileReport reasonerProfileReport = null;
            
            for(final OWLProfile nextInProfile : inProfiles)
            {
                if(nextInProfile.getIRI().equals(reasonerProfile))
                {
                    inReasonerProfile = true;
                }
            }
            
            for(final OWLProfileReport nextInProfileReport : inProfileReports)
            {
                if(nextInProfileReport.getProfile().getIRI().equals(reasonerProfile))
                {
                    reasonerProfileReport = nextInProfileReport;
                }
            }
            
            if(!inReasonerProfile)
            {
                this.log.error("Not using reasoner for inferencing as the configured reasoner profile was violated: ",
                        reasonerProfile);
                
                if(this.log.isTraceEnabled())
                {
                    this.log.trace("Reasoner profile violations: {}", reasonerProfileReport);
                }
            }
            else
            {
                final OWLReasonerFactory configuredReasoner =
                        OWLReasonerFactoryRegistry.getInstance().getReasonerFactory(reasonerName);
                
                if(configuredReasoner != null)
                {
                    if(this.log.isDebugEnabled())
                    {
                        this.log.debug("About to check for consistency using reasoner: {}",
                                configuredReasoner.getReasonerName());
                    }
                    
                    // TODO: May have to fork a thread off for this process if it takes too long and
                    // causes timeouts or otherwise unwanted delays
                    try
                    {
                        // Construct a random URI for the inferred axioms ontology
                        final String inferredOntologyUri =
                                "urn:oas:inferredontology:"
                                        + URLEncoder.encode(configuredReasoner.getReasonerName(), "UTF-8")
                                        + UUID.randomUUID().toString();
                        
                        OWLOntology inferredAxiomsOntology = null;
                        
                        inferredAxiomsOntology = this.manager.createOntology(IRI.create(inferredOntologyUri));
                        
                        final OWLReasoner reasoner = configuredReasoner.createReasoner(parsedOntology);
                        
                        if(!reasoner.isConsistent())
                        {
                            this.log.error("Ontology was not consistent when used with OWLReasoner.isConsistent");
                        }
                        else if(this.log.isDebugEnabled())
                        {
                            this.log.debug("Ontology was consistent when used with OWLReasoner.isConsistent");
                        }
                        
                        reasoner.precomputeInferences(InferenceType.CLASS_HIERARCHY);
                        
                        // TODO: How often will an explicit list of InferredAxiomGenerator instances
                        // be
                        // needed here?
                        // Create an InferredOntologyGenerator using the default list of
                        // InferredAxiomGenerators
                        final InferredOntologyGenerator iog = new InferredOntologyGenerator(reasoner);
                        
                        // the original parsedOntology is already known inside of the reasoner since
                        // the
                        // call to
                        iog.fillOntology(this.manager, inferredAxiomsOntology);
                        
                        nextOntology.setConcreteAxiomCount(parsedOntology.getAxiomCount());
                        nextOntology.setInferredAxiomCount(inferredAxiomsOntology.getAxiomCount());
                        
                        nextOntology.setConcreteLogicalAxiomCount(parsedOntology.getLogicalAxiomCount());
                        nextOntology.setIndividualCount(parsedOntology.getIndividualsInSignature(false).size());
                        
                        final RioRDFOntologyFormat nextFormat = new RDFXMLOntologyFormat();
                        
                        if(parsedFormat instanceof PrefixOWLOntologyFormat)
                        {
                            nextFormat.copyPrefixesFrom((PrefixOWLOntologyFormat)parsedFormat);
                        }
                        
                        final RioRenderer inferredAxiomsRenderer =
                                new RioRenderer(inferredAxiomsOntology, new StatementCollector(inferredModel),
                                        nextFormat);
                        inferredAxiomsRenderer.render();
                        
                        // if everything was processed properly then add the inferred ontology URI
                        // to the ontology.
                        // NOTE: If this step is not performed, then any inferred statements in
                        // inferredAxiomsStatementCollector will not be inserted into the database
                        nextOntology.setInferredOntologyUri(this.createURI(inferredOntologyUri));
                    }
                    catch(final OWLOntologyCreationException e)
                    {
                        this.log.error("Inferred Axioms Ontology was already in the OWLOntologyManager, ignoring it and not inferring anything.");
                    }
                    catch(final UnsupportedEncodingException e)
                    {
                        throw new ResourceException(
                                Status.SERVER_ERROR_INTERNAL,
                                "Virtual machine does not support UTF-8, abort now as there is no hope of consistency.",
                                e);
                    }
                    catch(final IOException e)
                    {
                        throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Failed to classify the ontology", e);
                    }
                    
                }
                else
                {
                    this.log.warn("Could not find configured reasoner: {}", reasonerName);
                }
            }
            
            final RioRDFOntologyFormat nextFormat = new RDFXMLOntologyFormat();
            
            if(parsedFormat instanceof PrefixOWLOntologyFormat)
            {
                nextFormat.copyPrefixesFrom((PrefixOWLOntologyFormat)parsedFormat);
            }
            
            final RioRenderer renderer =
                    new RioRenderer(parsedOntology, new StatementCollector(concreteModel), nextFormat);
            try
            {
                renderer.render();
            }
            catch(final IOException ioe)
            {
                this.log.error("Failed to complete storage, as an IO exception occurred.", ioe);
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                        "Failed to complete storage, as an IO exception occurred.", ioe);
            }
        }
    }
    
    /**
     * Helper method required to aggregate the Sesame functionality required by storeOntology
     * 
     * @param nextOntology
     * @param deleteOldVersions
     * @param concreteModel
     * @param inferredModel
     * @throws ResourceException
     */
    protected void storeOntologyInternalSesame(final Ontology nextOntology, final boolean deleteOldVersions,
            final Model concreteModel, final Model inferredModel) throws ResourceException
    {
        try
        {
            final List<Ontology> ontologiesByUri = this.getOntologiesByUri(nextOntology.getOntologyUri());
            
            if(ontologiesByUri.isEmpty())
            {
                if(nextOntology.getOntologyVersionUri() == null)
                {
                    nextOntology.setOntologyVersionUri(SesameUtils.generateNewOntologyVersionUri(
                            nextOntology.getOntologyUri(), null, this.repository, SesameUtils.ONTOLOGY_MANAGER_GRAPH));
                }
                
                // try to force generation of a new managed ontology version, without using an
                // existing version
                // managedOntologyVersionUri =
                // nextOntology.getManagedOntologyVersionUri(this.repository, true, true,
                // UriConstants.ONTOLOGY_MANAGER_GRAPH);
                nextOntology
                        .ensureVersionUriUniqueness(this.repository, true, true, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                
                if(nextOntology.getOntologyVersionUri() == null)
                {
                    this.log.warn("Could not find a managed ontology version URI for the ontology nextOntology.getOntologyUri()="
                            + nextOntology.getOntologyUri()
                            + " nextOntology.getOntologyVersionUri()="
                            + nextOntology.getOntologyVersionUri());
                }
            }
            else
            {
                final List<URI> currentVersionUris = new ArrayList<URI>(ontologiesByUri.size());
                
                for(final Ontology nextStoredOntology : ontologiesByUri)
                {
                    if(nextStoredOntology.getOntologyVersionUri() == null)
                    {
                        this.log.warn("Found fetched ontology with no versionURI nextStoredOntology.getOntologyUri()="
                                + nextStoredOntology.getOntologyUri());
                    }
                    else
                    {
                        // do not attempt to reuse existing version for this ontology
                        // managedOntologyVersionUri =
                        // nextStoredOntology.getManagedOntologyVersionUri(this.repository,
                        // false,
                        // true,
                        // UriConstants.ONTOLOGY_MANAGER_GRAPH);
                        
                        nextStoredOntology.ensureVersionUriUniqueness(this.repository, false, true,
                                SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                        
                        if(currentVersionUris.contains(nextOntology.getOntologyVersionUri()))
                        {
                            this.log.warn("Found duplicate version URI in fetched ontologies for a single ontology URI nextStoredOntology.getOntologyUri()="
                                    + nextStoredOntology.getOntologyUri());
                        }
                        else
                        {
                            currentVersionUris.add(nextStoredOntology.getOntologyVersionUri());
                        }
                    }
                }
                
                // if all fetched ontologies did not contain version URIs, then
                // currentVersionUris
                // will not receive any entries, so default back to our previous strategy in
                // this
                // case
                if(currentVersionUris.isEmpty())
                {
                    this.log.warn(
                            "There were NO existing ontology versions attached to this ontology URI, although the ontology was known: {}",
                            currentVersionUris);
                    
                    // setup the ontology version URI before setting up the managed version if
                    // there
                    // was no current version
                    if(nextOntology.getOntologyVersionUri() == null)
                    {
                        final URI newOntologyVersionUri =
                                SesameUtils.generateNewOntologyVersionUri(nextOntology.getOntologyUri(), null,
                                        this.repository, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                        
                        nextOntology.setOntologyVersionUri(newOntologyVersionUri);
                    }
                    
                    // do not attempt to use existing as we determined that there were no
                    // existing
                    // in this branch
                    // nextOntology.getManagedOntologyVersionUri(this.repository, false, true,
                    // UriConstants.ONTOLOGY_MANAGER_GRAPH);
                    nextOntology.ensureVersionUriUniqueness(this.repository, false, true,
                            SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                }
                // otherwise find a distinct new version URI, based on the discovered version
                // URIs
                else
                {
                    if(this.log.isDebugEnabled())
                    {
                        this.log.debug("There were existing ontology versions attached to this ontology URI: {}",
                                currentVersionUris);
                    }
                    
                    // sort the list, and choose the last one to modify it
                    // NOTE: This guarantees that the generated URI will be distinct if it is
                    // designed to sort after all other URIs using the sesametools
                    // ValueComparator,
                    // which I designed
                    
                    // IMPORTANT: The comparator requires
                    // currentVersionUri.stringValue().compare(newDistinctVersionUri.stringValue())
                    // to be less than 0 for the current version IRI string value
                    // If this contract is not followed, then getManagedOntologyVersionUri will
                    // return a different value
                    Collections.sort(currentVersionUris, new ValueComparator());
                    
                    // If there was a version for this ontology in the database but the user did
                    // not
                    // specify the version, then we generate a new ontology version Uri for it
                    if(nextOntology.getOntologyVersionUri() == null)
                    {
                        if(this.log.isDebugEnabled())
                        {
                            this.log.debug("Ontology version was not specified by the user, and there was a previous version in the database, generating new version IRI");
                        }
                        
                        nextOntology.setOntologyVersionUri(SesameUtils.generateNewOntologyVersionUri(
                                nextOntology.getOntologyUri(), currentVersionUris.get(currentVersionUris.size() - 1),
                                this.repository, SesameUtils.ONTOLOGY_MANAGER_GRAPH));
                    }
                    
                    // find a managed ontology version URI, without an existing version
                    // managedOntologyVersionUri =
                    // nextOntology.getManagedOntologyVersionUri(this.repository, true, false,
                    // UriConstants.ONTOLOGY_MANAGER_GRAPH);
                    
                    nextOntology.ensureVersionUriUniqueness(this.repository, true, false,
                            SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                    
                    if(this.log.isDebugEnabled())
                    {
                        this.log.debug("managedOntologyVersionUri={}", nextOntology.getOntologyVersionUri());
                    }
                    
                    // if the version was found to be a duplicate then generate a new version
                    // URI
                    // and regenerate the managed ontology version URI afterwards
                    // TODO: Iterate here until a unique version is obtained
                    if(currentVersionUris.contains(nextOntology.getOntologyVersionUri()))
                    {
                        if(this.log.isDebugEnabled())
                        {
                            this.log.debug("The managedOntologyVersionUri was already known about, so attempting to generate a new one");
                        }
                        
                        nextOntology.setOntologyVersionUri(SesameUtils.generateNewOntologyVersionUri(
                                nextOntology.getOntologyUri(), currentVersionUris.get(currentVersionUris.size() - 1),
                                this.repository, SesameUtils.ONTOLOGY_MANAGER_GRAPH));
                        
                        // refetch the managed ontology version after setting the ontology
                        // version
                        // URI to a new URI
                        // NOTE: Do not use existing versions at this point as we determined
                        // that
                        // there was an overlap/duplicate
                        // managedOntologyVersionUri =
                        // nextOntology.getManagedOntologyVersionUri(this.repository, true,
                        // false,
                        // UriConstants.ONTOLOGY_MANAGER_GRAPH);
                        nextOntology.ensureVersionUriUniqueness(this.repository, true, false,
                                SesameUtils.ONTOLOGY_MANAGER_GRAPH);
                        
                        if(this.log.isDebugEnabled())
                        {
                            this.log.debug("The last resort generated managedOntologyVersionUri is: {}",
                                    nextOntology.getOntologyVersionUri());
                        }
                        
                        if(currentVersionUris.contains(nextOntology.getOntologyVersionUri()))
                        {
                            this.log.error("getManagedOntologyVersionUri failed to generate a unique result");
                            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                                    "Failed to complete storage, as a new unique version IRI was not able to be generated.");
                        }
                    }
                    
                }
            }
            
            // Important that we attempt to generate the new ontology version while the other
            // ontologies are in the repository, even if deleteOldVersions parameter to this
            // method was true
            // If we wait to generate it later, it may have already been retrieved in the past
            // and documented somewhere as a version, which would not be the same as the current
            // version
            if(deleteOldVersions)
            {
                if(this.log.isInfoEnabled())
                {
                    this.log.info("Delete old versions was true");
                    
                    for(final Ontology nextDeletionOntology : ontologiesByUri)
                    {
                        this.log.info("Deleting ontology: {}", nextDeletionOntology);
                        
                        this.deleteOntologyVersion(nextDeletionOntology);
                    }
                }
            }
            
            SesameUtils.storeOntology(nextOntology, concreteModel, Namespaces.asMap(concreteModel.getNamespaces()),
                    inferredModel, this.repository, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
            
        }
        catch(final RepositoryException re)
        {
            this.log.error("Failed to complete storage, as a repository exception occurred.", re);
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete storage, as a repository exception occurred.", re);
        }
        catch(final RDFParseException rdfpe)
        {
            this.log.error("Failed to complete storage, as an RDF parse exception occurred.", rdfpe);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                    "Failed to complete storage, as an RDF parse exception occurred.", rdfpe);
        }
        catch(final IOException ioe)
        {
            this.log.error("Failed to complete storage, as an IO exception occurred.", ioe);
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete storage, as an IO exception occurred.", ioe);
        }
        catch(final RDFHandlerException rdfhe)
        {
            this.log.error("Failed to complete storage, as a repository exception occurred.", rdfhe);
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete storage, as a repository exception occurred.", rdfhe);
        }
    }
    
    @Override
    public boolean updateAnnotation(final Annotation nextAnnotation)
    {
        this.log.info("Updating annotation: {}", nextAnnotation.getAnnotationUniqueId().stringValue());
        
        // this.addLogDetailsForRequest(nextAnnotation.getAnnotationUniqueId().stringValue(),
        // request, DataHandlerConstants.UPDATE);
        
        // TODO: what should we do if the delete failed
        final boolean deleteResult = this.deleteAnnotation(nextAnnotation.getAnnotationUniqueId());
        
        if(!deleteResult)
        {
            throw new IllegalStateException("Could not update annotation as it did not previously exist");
        }
        
        final boolean storeResult = this.storeAnnotation(nextAnnotation);
        
        return storeResult;
    }
}
