/**
 * 
 */
package com.github.ansell.oas.objects.impl;

import org.openrdf.model.URI;

import com.github.ansell.oas.objects.ResourceCount;

/**
 * An implementation of the ResourceCount interface that can be overriden to provide counts specific
 * to particular circumstances.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public abstract class AbstractResourceCountImpl implements ResourceCount
{
    protected int counter = 0;
    private URI resource;
    
    /**
     * Default constructor.
     */
    public AbstractResourceCountImpl()
    {
        this(0);
    }
    
    /**
     * Constructor to initialise count to a particular value.
     * 
     * @param count
     *            The initial count to use.
     */
    public AbstractResourceCountImpl(final int count)
    {
        this.setCount(count);
    }
    
    @Override
    public int accumulate(final int increment)
    {
        this.counter += increment;
        
        return this.counter;
    }
    
    @Override
    public int getCount()
    {
        return this.counter;
    }
    
    @Override
    public void setCount(final int count)
    {
        this.counter = count;
    }
    
    @Override
    public URI getResourceURI()
    {
        return this.resource;
    }
    
    @Override
    public void setResourceURI(URI resource)
    {
        this.resource = resource;
    }
}
