/**
 * 
 */
package com.github.ansell.oas.utils;

import java.util.Map;

import org.restlet.Request;
import org.restlet.Response;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.service.StatusService;

import com.github.ansell.propertyutil.PropertyUtil;

import freemarker.template.Configuration;

/**
 * This status service is based on the standard Restlet guide for returning custom error pages.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class OasStatusService extends StatusService
{
    
    private Configuration freemarkerConfiguration;
    private PropertyUtil propertyUtil;
    
    /**
     * @param enabled
     */
    public OasStatusService(final Configuration freemarkerConfiguration, final boolean enabled)
    {
        super(enabled);
        this.freemarkerConfiguration = freemarkerConfiguration;
    }
    
    /**
     * 
     */
    public OasStatusService(final Configuration freemarkerConfiguration, final PropertyUtil propertyUtil)
    {
        super();
        this.freemarkerConfiguration = freemarkerConfiguration;
        this.propertyUtil = propertyUtil;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see org.restlet.service.StatusService#getRepresentation(org.restlet.data.Status,
     * org.restlet.Request, org.restlet.Response)
     */
    @Override
    public Representation getRepresentation(final Status status, final Request request, final Response response)
    {
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(request);
        
        dataModel.put("contentTemplate", this.propertyUtil.get(OasProps.PROP_TMPL_ERROR, OasProps.DEF_TMPL_ERROR));
        dataModel.put("pageTitle", "An error occurred : HTTP " + status.getCode());
        
        if(status.getThrowable() != null)
        {
            dataModel.put("errorMessage", status.getThrowable().getMessage());
        }
        else
        {
            dataModel.put("errorMessage", status.getDescription());
        }
        // TODO: Support non-HTML error representations
        return RestletUtils.getHtmlRepresentation(
                this.propertyUtil.get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel, MediaType.TEXT_HTML,
                this.freemarkerConfiguration);
        
    }
    
}
