/**
 * 
 */
package com.github.ansell.oas.utils;

import org.openrdf.model.URI;
import org.openrdf.repository.RepositoryException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.restletutils.RestletUtilSesameRealm;
import com.github.ansell.restletutils.RestletUtilUser;

/**
 * Used to translate user identifier URIs to human readable strings in freemarker templates.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public final class UserUtils
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private final RestletUtilSesameRealm realm;
    
    /**
     * 
     */
    public UserUtils(final RestletUtilSesameRealm realm)
    {
        this.realm = realm;
    }
    
    public String displayname(final URI userUri)
    {
        final String username = this.username(userUri);
        
        if(username.startsWith("Unknown user"))
        {
            return username;
        }
        
        return this.displaynameforusername(username);
    }
    
    public String displaynameforusername(final String username)
    {
        if(username.startsWith("Unknown user"))
        {
            return username;
        }
        
        try
        {
            final RestletUtilUser user = this.realm.findUser(username);
            
            if(user != null)
            {
                return user.getFirstName() + " " + user.getLastName();
            }
        }
        catch(final RuntimeException e)
        {
            this.log.error("Repository exception while searching for user for username", e);
        }
        
        return username;
    }
    
    /**
     * 
     * @param userIdentifier
     *            The URI of a user.
     * @return A human readable label for the given identifier.
     */
    public String username(final URI userUri)
    {
        try
        {
            if(userUri != null)
            {
                final String username = this.realm.getUsername(userUri);
                
                if(username != null)
                {
                    return username;
                }
                else
                {
                    return "Unknown user: " + userUri.stringValue();
                }
            }
        }
        catch(final RepositoryException e)
        {
            this.log.error("Repository exception while searching for username for user URI", e);
        }
        
        return "Unknown user";
    }
}
