/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.util.Collection;
import java.util.Map;

import org.openrdf.model.URI;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFHandlerException;
import org.restlet.data.CharacterSet;
import org.restlet.data.Form;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.objects.ResourceCount;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.utils.UserUtils;
import com.github.ansell.oas.webservice.AnnotationCountsByAnnotatedObjectUriResource;
import com.github.ansell.oas.webservice.BaseOasActions;

/**
 * Implementation of the webservice for fetching annotations based on the URI of the annotated
 * object target.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class AnnotationCountsByAnnotatedObjectUriResourceImpl extends AbstractOasResourceImpl implements
        AnnotationCountsByAnnotatedObjectUriResource
{
    public AnnotationCountsByAnnotatedObjectUriResourceImpl()
    {
        super();
    }
    
    @Override
    public Representation getHtmlAnnotationCountByAnnotatedObjectUri(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final Form queryParams = this.getReference().getQueryAsForm(CharacterSet.UTF_8);
        
        // get the query parameter, ignoring the case, so ObjectURI and objectURI will both match
        // this
        final String objectUri = queryParams.getFirstValue("objecturi", true);
        
        if(objectUri == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "ObjectUri query parameter was not present");
        }
        
        final URI objectUriResource = this.getDataHandler().createURI(objectUri);
        
        final Collection<Annotation> results =
                this.getDataHandler().getAnnotationsByAnnotatedObjectUri(objectUriResource);
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST, OasProps.DEF_TMPL_ANNOTATIONS_LIST));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST_TITLE,
                        OasProps.DEF_TMPL_ANNOTATIONS_LIST_TITLE));
        
        dataModel.put("annotations", results);
        dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        
        return RestletUtils.getHtmlRepresentation(
                this.getOasApplication().getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE),
                dataModel, MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    public Representation getRdfAnnotationCountByAnnotatedObjectUri(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final Form queryParams = this.getReference().getQueryAsForm(CharacterSet.UTF_8);
        // Parameter fullHtml = queryParams.getFirst("fullhtml", true);
        //
        // boolean returnFullHtml = RestletUtils.getBooleanFromParameter(fullHtml);
        
        // get the query parameter, ignoring the case, so ObjectURI and objectURI will both match
        // this
        final String objectUri = queryParams.getFirstValue("objecturi", true);
        
        if(objectUri == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "ObjectUri query parameter was not present");
        }
        
        final URI objectUriResource = this.getDataHandler().createURI(objectUri);
        
        final ResourceCount results = this.getDataHandler().getAnnotationCountsByAnnotatedObjectUri(objectUriResource);
        
        try
        {
            return RestletUtils.getRdfRepresentation(results, variant, this.getRequest().getResourceRef());
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final RDFHandlerException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
    }
}
