/**
 * Package for implementations of the resource handlers in the OAS API
 */
package com.github.ansell.oas.webservice.impl;

import org.openrdf.model.URI;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.BulkDeleteAnnotationResource;

/**
 * Creates, fetches and updates annotations.
 * 
 * Provides responses in HTML, Talis RDF/JSON and RDF/XML.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class BulkDeleteAnnotationResourceImpl extends AbstractOasResourceImpl implements BulkDeleteAnnotationResource
{
    /**
     * Default constructor for Annotation Resource Implementation.
     */
    public BulkDeleteAnnotationResourceImpl()
    {
        super();
    }
    
    @Override
    public void deleteAnnotations(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_DELETE);
        
        final String inputAnnotationId =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue(OasProps.PARAM_ANNOTATIONID, true);
        
        if(inputAnnotationId == null)
        {
            this.log.error("No annotation specified in request");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "No annotation specified.");
        }
        
        try
        {
            final URI annotationId = this.getDataHandler().createURI(inputAnnotationId);
            
            // Get the annotation based on the identifier URI and load it into memory
            final Annotation annotation = this.getDataHandler().getAnnotation(annotationId);
            
            if(annotation == null)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find annotation to delete");
            }
            
            if(this.getDataHandler().deleteAnnotation(annotationId))
            {
                this.redirectToPath(this.getPropertyUtil().get(OasProps.PROP_BULK_FETCH_ANNOTATION_PATH,
                        OasProps.DEF_BULK_FETCH_ANNOTATION_PATH));
                // HTTP 204 No Content if the delete succeeded
                // this.getResponse().setStatus(Status.SUCCESS_NO_CONTENT);
            }
            else
            {
                // if there were no exceptions, but the delete failed, assume it was the clients
                // fault
                // HTTP Status.CLIENT_ERROR_BAD_REQUEST Client Error if the delete did not succeed
                // but
                // there were no exceptions
                // thrown
                this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST, "Delete annotation failed");
            }
        }
        catch(final IllegalArgumentException iae)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid annotation ID");
        }
        
    }
    
}
