/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.util.Map;

import org.restlet.data.MediaType;
import org.restlet.data.Reference;
import org.restlet.representation.Representation;
import org.restlet.resource.Get;
import org.restlet.resource.ResourceException;
import org.restlet.security.User;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.LoginResource;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class CookieLoginResourceImpl extends AbstractOasResourceImpl implements LoginResource
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    /**
     * 
     */
    public CookieLoginResourceImpl()
    {
        super();
    }
    
    private String getDefaultRedirect()
    {
        return this.getRequest().getRootRef().toString()
                + this.getPropertyUtil().get(OasProps.PROP_INDEX_PATH, OasProps.DEF_INDEX_PATH);
    }
    
    @Override
    @Get("html")
    public Representation getLoginPageHtml(final Representation entity) throws ResourceException
    {
        final User user = this.getRequest().getClientInfo().getUser();
        
        this.log.debug("authenticated user: {}", user);
        
        if(user == null)
        {
            this.log.debug("In getLoginPageHtml");
            final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
            dataModel.put("contentTemplate",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_LOGIN, OasProps.DEF_TMPL_LOGIN));
            dataModel.put("pageTitle",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_LOGIN_TITLE, OasProps.DEF_TMPL_LOGIN_TITLE));
            
            String referrerRef = this.getDefaultRedirect();
            
            // By default use the referrer to populate the redirectTo field internally for
            // use after a successful login, unless the request contains a referrerRef, in which
            // case redirect to it
            if(this.getRequest().getReferrerRef() != null)
            {
                referrerRef = this.getRequest().getReferrerRef().toString(true, true);
            }
            
            dataModel.put("referrerRef", referrerRef);
            this.log.debug("referrerRef={}", referrerRef);
            
            // Output the base template, with contentTemplate from the dataModel defining the
            // template
            // to use for the content in the body of the page
            return RestletUtils.getHtmlRepresentation(
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                    MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
        }
        else
        {
            // Bounce users back to their referrer if they are already logged in, or if they did not
            // send a referrer use the default redirect instead
            final Reference referrerRef = this.getRequest().getReferrerRef();
            
            this.log.debug("referrerRef={}", referrerRef);
            
            if(referrerRef != null && !referrerRef.equals(this.getRequest().getResourceRef()))
            {
                this.getResponse().redirectSeeOther(referrerRef);
            }
            else
            {
                this.getResponse().redirectSeeOther(this.getDefaultRedirect());
            }
            return null;
        }
    }
    
}
