/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.resource.Get;
import org.restlet.resource.ResourceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class HelpResourceImpl extends AbstractOasResourceImpl
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    @Get("html")
    public Representation getHelpPageHtml(final Representation entity) throws ResourceException
    {
        this.log.debug("in getHelpPageHtml");
        
        // String helpPageString =
        // this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("page", true);
        
        String helpPageString = (String)this.getRequest().getAttributes().get("page");
        
        if(helpPageString == null || helpPageString.trim().isEmpty())
        {
            this.log.error("Found not find request attribute, so defauling to 'navigation'");
            helpPageString = "navigation";
        }
        
        helpPageString = helpPageString.trim();
        
        if(!StringUtils.isAlphanumeric(helpPageString))
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not retrieve the requested help page");
        }
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_HELP_BASE, OasProps.DEF_TMPL_HELP_BASE) + helpPageString
                        + ".html.ftl");
        dataModel.put("pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_INDEX_TITLE, OasProps.DEF_TMPL_INDEX_TITLE));
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
}
