/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import org.restlet.data.Status;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.webservice.SetCurrentOntologyResource;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class SetCurrentOntologyResourceImpl extends AbstractOasResourceImpl implements SetCurrentOntologyResource
{
    @Override
    public void setCurrentOntology() throws ResourceException
    {
        final String ontologyUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        this.log.debug("ontologyUriString={}", ontologyUriString);
        
        if(ontologyUriString == null)
        {
            this.log.error("No ontology URI specified");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "No ontology URI specified in request");
        }
        
        final String versionUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyVersionUri", true);
        
        this.log.debug("versionUriString={}", versionUriString);
        
        if(versionUriString == null)
        {
            this.log.error("No ontology versuib URI specified");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "No ontology version URI specified in request");
        }
        
        Ontology latestOntology =
                this.getDataHandler().setLatestOntology(this.getDataHandler().createURI(ontologyUriString),
                        this.getDataHandler().createURI(versionUriString));
        
        this.redirectToPath(this.getPropertyUtil().get(OasProps.PROP_ONTOLOGY_MANAGER_PATH,
                OasProps.DEF_ONTOLOGY_MANAGER_PATH)
                + "?ontologyUri="
                + latestOntology.getOntologyUri().stringValue()
                + "&ontologyVersionUri="
                + latestOntology.getOntologyVersionUri().stringValue());
    }
    
}
